# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import os
import random
import sys
import threading
from collections import namedtuple, defaultdict
from enum import Enum
from typing import List


_context_factory = threading.local()


def get_context() -> "ContextBase":
    return getattr(_context_factory, 'current', None)


class RunningMode(Enum):
    local = 0
    local_cluster = 1
    distributed = 2


class ContextBase(object):
    """
    Context will be used as a global object to detect the environment,
    and fetch meta, etc, mostly used in the server side, not for user.
    """

    @property
    def running_mode(self):
        """
        Get the running mode, could be local, local_cluster or distributed.
        """
        raise NotImplementedError

    @property
    def session_id(self):
        """
        Get session id.
        """
        raise NotImplementedError

    def __enter__(self):
        _context_factory.prev = getattr(_context_factory, 'current', None)
        _context_factory.current = self

    def __exit__(self, *_):
        _context_factory.current = _context_factory.prev
        _context_factory.prev = None

    def wraps(self, func):
        def h(*args, **kwargs):
            with self:
                return func(*args, **kwargs)
        return h

    def get_current_session(self):
        """
        Get current session.

        :return: Session
        """
        raise NotImplementedError

    # ------------
    # Meta related
    # ------------

    def get_tileable_metas(self, tileable_keys, filter_fields: List[str] = None) -> List:
        """
        get tileable metas. Tileable includes tensor, DataFrame, mutable tensor and mutable DataFrame.
        :param tileable_keys: tileable keys
        :param filter_fields: filter the fields in meta
        :return: List of tileable metas
        """
        raise NotImplementedError

    def get_chunk_metas(self, chunk_keys, filter_fields=None):
        """
        Get chunk metas according to the given chunk keys.

        :param chunk_keys: chunk keys
        :param filter_fields: filter the fields in meta
        :return: List of chunk metas
        """
        raise NotImplementedError

    # ---------------
    # Cluster related
    # ---------------

    def get_scheduler_addresses(self):
        """
        Get scheduler addresses

        :return: List of scheduler addresses
        """
        raise NotImplementedError

    def get_worker_addresses(self):
        """
        Get worker addreses

        :return: List of worker addresses
        """
        raise NotImplementedError

    def get_worker_metas(self):  # pragma: no cover
        """
        Get worker metas

        :return: List of worker metas
        """
        raise NotImplementedError

    def get_local_address(self):
        """
        Get local address

        :return: local address
        """
        raise NotImplementedError

    def get_ncores(self):
        """
        Get number of cores
        """
        raise NotImplementedError

    # -------------
    # Graph related
    # -------------

    def submit_chunk_graph(self, graph, result_keys):
        """
        Submit fine-grained graph to execute.

        :param graph: fine-grained graph to execute
        :param result_keys: result chunk keys
        :return: Future
        """
        raise NotImplementedError

    def submit_tileable_graph(self, graph, result_keys):
        """
        Submit coarse-grained graph to execute.

        :param graph: coarse-grained graph to execute
        :param result_keys: result tileable keys
        :return: Future
        """
        raise NotImplementedError

    # -----------------------
    # Pool occupation related
    # -----------------------

    def yield_execution_pool(self):
        """
        Yields current execution pool to allow running other tasks
        """
        pass

    def acquire_execution_pool(self, yield_info):
        pass

    # --------------
    # Result related
    # --------------

    def get_chunk_results(self, chunk_keys: List[str]) -> List:
        """
        Get results when given chunk keys

        :param chunk_keys: chunk keys
        :return: list of chunk results
        """
        raise NotImplementedError

    # ------
    # Others
    # ------

    def create_lock(self):
        raise NotImplementedError

    def get_named_tileable_infos(self, name: str):
        raise NotImplementedError


ChunkMeta = namedtuple('ChunkMeta', ['chunk_size', 'chunk_shape', 'workers'])
TileableInfos = namedtuple('TileableInfos', ['tileable_key', 'tileable_shape'])


class LocalContext(ContextBase, dict):
    def __init__(self, local_session, ncores=None):
        dict.__init__(self)
        self._local_session = local_session
        self._ncores = ncores

    def copy(self):
        new_d = LocalContext(self._local_session, ncores=self._ncores)
        new_d.update(self)
        return new_d

    def get_current_session(self):
        from .session import new_session

        sess = new_session()
        sess._sess = copy.copy(self._local_session)
        sess._sess.context = self
        return sess

    def set_ncores(self, ncores):
        self._ncores = ncores

    @property
    def running_mode(self):
        return RunningMode.local

    @property
    def session_id(self):
        return self._local_session.session_id

    def get_scheduler_addresses(self):
        return

    def get_worker_addresses(self):  # pragma: no cover
        return

    def get_worker_metas(self):  # pragma: no cover
        return

    def get_local_address(self):
        return

    def get_ncores(self):
        return self._ncores

    def get_chunk_metas(self, chunk_keys, filter_fields=None):
        metas = []
        for chunk_key in chunk_keys:
            chunk_data = self.get(chunk_key)
            if chunk_data is None:
                metas.append(None)
                continue
            if hasattr(chunk_data, 'nbytes'):
                # ndarray
                size = chunk_data.nbytes
                shape = chunk_data.shape
            elif hasattr(chunk_data, 'memory_usage'):
                # DataFrame
                size = chunk_data.memory_usage(deep=True).sum()
                shape = chunk_data.shape
            else:
                # other
                size = sys.getsizeof(chunk_data)
                shape = ()

            metas.append(ChunkMeta(chunk_size=size, chunk_shape=shape, workers=None))

        selected_metas = []
        for chunk_meta in metas:
            if filter_fields is not None:
                chunk_meta = [getattr(chunk_meta, field) for field in filter_fields]
            selected_metas.append(chunk_meta)
        return selected_metas

    def get_chunk_results(self, chunk_keys: List[str]) -> List:
        # As the context is actually holding the data,
        # so for the local context, we just fetch data from itself
        return [self[chunk_key] for chunk_key in chunk_keys]

    def get_named_tileable_infos(self, name: str):
        if name not in self._local_session.executor._tileable_names:
            raise ValueError(f"Name {name} doesn't exist.")
        tileable = self._local_session.executor._tileable_names[name]
        return TileableInfos(tileable.key, tileable.shape)

    def create_lock(self):
        return self._local_session.executor._sync_provider.lock()


class DistributedContext(ContextBase):
    def __init__(self, scheduler_address, session_id, actor_ctx=None, **kw):
        from .worker.api import WorkerAPI
        from .scheduler.resource import ResourceActor
        from .scheduler.utils import SchedulerClusterInfoActor
        from .actors import new_client

        self._session_id = session_id
        self._scheduler_address = scheduler_address
        self._worker_api = WorkerAPI()
        self._meta_api_thread_local = threading.local()

        self._running_mode = None
        self._actor_ctx = actor_ctx or new_client()
        self._cluster_info = self._actor_ctx.actor_ref(
            SchedulerClusterInfoActor.default_uid(), address=scheduler_address)
        is_distributed = self._cluster_info.is_distributed()
        self._running_mode = RunningMode.local_cluster \
            if not is_distributed else RunningMode.distributed
        self._resource_actor_ref = self._actor_ctx.actor_ref(
            ResourceActor.default_uid(), address=scheduler_address)

        self._address = kw.pop('address', None)
        self._extra_info = kw

    @property
    def meta_api(self):
        from .scheduler.api import MetaAPI
        try:
            return self._meta_api_thread_local._meta_api
        except AttributeError:
            meta_api = self._meta_api_thread_local._meta_api \
                = MetaAPI(scheduler_endpoint=self._scheduler_address)
            return meta_api

    @property
    def running_mode(self):
        return self._running_mode

    @property
    def session_id(self):
        return self._session_id

    def get_current_session(self):
        from .session import new_session, ClusterSession

        sess = new_session()
        sess._sess = ClusterSession(self._scheduler_address,
                                    session_id=self._session_id)
        return sess

    def get_scheduler_addresses(self):
        return self._cluster_info.get_schedulers()

    def get_worker_addresses(self):
        return self._resource_actor_ref.get_worker_endpoints()

    def get_worker_metas(self):  # pragma: no cover
        return self._resource_actor_ref.get_worker_metas()

    def get_local_address(self):
        return self._address

    def get_ncores(self):
        return self._extra_info.get('n_cpu')

    def get_chunk_results(self, chunk_keys: List[str]) -> List:
        from .serialize import dataserializer
        from .worker.transfer import ResultSenderActor

        all_workers = [m.workers for m in self.get_chunk_metas(chunk_keys)]
        results = []
        for chunk_key, endpoints in zip(chunk_keys, all_workers):
            sender_ref = self._actor_ctx.actor_ref(
                ResultSenderActor.default_uid(), address=endpoints[-1])
            results.append(
                dataserializer.loads(sender_ref.fetch_data(self._session_id, chunk_key)))
        return results

    # Meta API
    def get_tileable_metas(self, tileable_keys, filter_fields: List[str] = None) -> List:
        return self.meta_api.get_tileable_metas(self._session_id, tileable_keys, filter_fields)

    def get_chunk_metas(self, chunk_keys, filter_fields: List[str] = None) -> List:
        return self.meta_api.get_chunk_metas(self._session_id, chunk_keys, filter_fields)

    def get_named_tileable_infos(self, name: str):
        tileable_key = self.meta_api.get_tileable_key_by_name(self._session_id, name)
        nsplits = self.get_tileable_metas([tileable_key], filter_fields=['nsplits'])[0][0]
        shape = tuple(sum(s) for s in nsplits)
        return TileableInfos(tileable_key, shape)

    # Worker API
    def get_chunks_data(self, worker: str, chunk_keys: List[str], indexes: List = None,
                        compression_types: List[str] = None):
        return self._worker_api.get_chunks_data(self._session_id, worker, chunk_keys, indexes=indexes,
                                                compression_types=compression_types)

    # Fetch tileable data by tileable keys and indexes.
    def get_tileable_data(self, tileable_key: str, indexes: List = None,
                          compression_types: List[str] = None):
        from .serialize import dataserializer
        from .utils import merge_chunks
        from .tensor.datasource import empty
        from .tensor.indexing.index_lib import NDArrayIndexesHandler

        nsplits, chunk_keys, chunk_indexes = self.get_tileable_metas([tileable_key])[0]
        chunk_idx_to_keys = dict(zip(chunk_indexes, chunk_keys))
        chunk_keys_to_idx = dict(zip(chunk_keys, chunk_indexes))
        endpoints = self.get_chunk_metas(chunk_keys, filter_fields=['workers'])
        chunk_keys_to_worker = dict((chunk_key, random.choice(es[0])) for es, chunk_key in zip(endpoints, chunk_keys))

        chunk_workers = defaultdict(list)
        [chunk_workers[e].append(chunk_key) for chunk_key, e in chunk_keys_to_worker.items()]

        chunk_results = dict()
        if indexes is None or len(indexes) == 0:
            datas = []
            for endpoint, chunks in chunk_workers.items():
                datas.append(self.get_chunks_data(endpoint, chunks, compression_types=compression_types))
            datas = [d.result() for d in datas]
            for (endpoint, chunks), d in zip(chunk_workers.items(), datas):
                d = [dataserializer.loads(db) for db in d]
                chunk_results.update(dict(zip([chunk_keys_to_idx[k] for k in chunks], d)))

            chunk_results = [(k, v) for k, v in chunk_results.items()]
            if len(chunk_results) == 1:
                return chunk_results[0][1]
            else:
                return merge_chunks(chunk_results)
        else:
            # Reuse the getitem logic to get each chunk's indexes
            tileable_shape = tuple(sum(s) for s in nsplits)
            empty_tileable = empty(tileable_shape, chunk_size=nsplits)._inplace_tile()
            indexed = empty_tileable[tuple(indexes)]
            indexes_handler = NDArrayIndexesHandler()
            try:
                context = indexes_handler.handle(indexed.op, return_context=True)
            except TypeError:
                raise TypeError("Doesn't support indexing by tensors")

            result_chunks = dict()
            for c in context.processed_chunks:
                result_chunks[chunk_idx_to_keys[c.inputs[0].index]] = [c.index, c.op.indexes]

            chunk_datas = dict()
            for endpoint, chunks in chunk_workers.items():
                to_fetch_keys = []
                to_fetch_indexes = []
                to_fetch_idx = []
                for r_chunk, (chunk_index, index_obj) in result_chunks.items():
                    if r_chunk in chunks:
                        to_fetch_keys.append(r_chunk)
                        to_fetch_indexes.append(index_obj)
                        to_fetch_idx.append(chunk_index)
                if to_fetch_keys:
                    datas = self.get_chunks_data(endpoint, to_fetch_keys, indexes=to_fetch_indexes,
                                                 compression_types=compression_types)
                    chunk_datas[tuple(to_fetch_idx)] = datas
            chunk_datas = dict((k, v.result()) for k, v in chunk_datas.items())
            for idx, d in chunk_datas.items():
                d = [dataserializer.loads(db) for db in d]
                chunk_results.update(dict(zip(idx, d)))

            chunk_results = [(k, v) for k, v in chunk_results.items()]
            return indexes_handler.aggregate_result(context, chunk_results)

    def create_lock(self):
        return self._actor_ctx.lock()


class DistributedDictContext(DistributedContext, dict):
    def __init__(self, *args, **kwargs):
        DistributedContext.__init__(self, *args, **kwargs)
        dict.__init__(self)

    def yield_execution_pool(self):
        actor_cls = self.get('_actor_cls')
        actor_uid = self.get('_actor_uid')
        op_key = self.get('_op_key')
        if not actor_cls or not actor_uid:  # pragma: no cover
            return

        from .actors import new_client
        from .worker.daemon import WorkerDaemonActor
        client = new_client()

        worker_addr = self.get_local_address()
        if client.has_actor(client.actor_ref(WorkerDaemonActor.default_uid(), address=worker_addr)):
            holder = client.actor_ref(WorkerDaemonActor.default_uid(), address=worker_addr)
        else:
            holder = client
        uid = f'w:0:mars-cpu-calc-backup-{os.getpid()}-{op_key}-{random.randint(-1, 9999)}'
        uid = self._actor_ctx.distributor.make_same_process(uid, actor_uid)
        ref = holder.create_actor(actor_cls, uid=uid, address=worker_addr)
        return ref

    def acquire_execution_pool(self, yield_info):
        if yield_info is None:
            return

        from .actors import new_client
        client = new_client()
        calc_ref = client.actor_ref(yield_info, address=self.get_local_address())
        calc_ref.mark_destroy()



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from ...actors import FunctionActor
from ...errors import StorageFull, StorageDataExists, SerializationFailed
from ...serialize import dataserializer
from ...utils import lazy_import

try:
    import pyarrow

    try:
        from pyarrow.serialization import SerializationCallbackError
    except ImportError:
        from pyarrow.lib import SerializationCallbackError
except ImportError:  # pragma: no cover
    pyarrow = None
    SerializationCallbackError = None

plasma = lazy_import('pyarrow.plasma', globals=globals(), rename='plasma')

logger = logging.getLogger(__name__)
PAGE_SIZE = 64 * 1024


class PlasmaErrors:
    def __getattr__(self, item):
        import pyarrow.lib

        ret = getattr(plasma, item, None)
        if ret is None:  # pragma: no cover
            if item == 'PlasmaObjectNotFound':
                ret = getattr(plasma, 'PlasmaObjectNonexistent', None) \
                      or getattr(pyarrow.lib, 'PlasmaObjectNonexistent')
            elif item == 'PlasmaStoreFull':
                ret = getattr(pyarrow.lib, item)

        if ret is not None:
            setattr(self, item, ret)
        return ret


plasma_errors = PlasmaErrors()


class PlasmaKeyMapActor(FunctionActor):
    @classmethod
    def default_uid(cls):
        return 'w:0:' + cls.__name__

    def __init__(self):
        super().__init__()
        self._mapping = dict()

    def put(self, session_id, chunk_key, obj_id):
        session_chunk_key = (session_id, chunk_key)
        if session_chunk_key in self._mapping:
            raise StorageDataExists(session_chunk_key)
        self._mapping[session_chunk_key] = obj_id

    def get(self, session_id, chunk_key):
        return self._mapping.get((session_id, chunk_key))

    def delete(self, session_id, chunk_key):
        try:
            del self._mapping[(session_id, chunk_key)]
        except KeyError:
            pass

    def batch_delete(self, session_id, chunk_keys):
        for k in chunk_keys:
            self.delete(session_id, k)


class PlasmaSharedStore(object):
    """
    Wrapper of plasma client for Mars objects
    """
    def __init__(self, plasma_client, mapper_ref):
        from ...serialize.dataserializer import mars_serialize_context

        self._plasma_client = plasma_client
        self._size_limit = None
        self._serialize_context = mars_serialize_context()

        self._mapper_ref = mapper_ref
        self._pool = mapper_ref.ctx.threadpool(1)

    def get_actual_capacity(self, store_limit):
        """
        Get actual capacity of plasma store
        :return: actual storage size in bytes
        """
        try:
            store_limit = min(store_limit, self._plasma_client.store_capacity())
        except AttributeError:  # pragma: no cover
            pass

        if self._size_limit is None:
            left_size = store_limit
            alloc_fraction = 1
            while True:
                allocate_size = int(left_size * alloc_fraction / PAGE_SIZE) * PAGE_SIZE
                try:
                    obj_id = plasma.ObjectID.from_random()
                    buf = [self._plasma_client.create(obj_id, allocate_size)]
                    self._plasma_client.seal(obj_id)
                    del buf[:]
                    break
                except plasma_errors.PlasmaStoreFull:
                    alloc_fraction *= 0.99
                finally:
                    self._plasma_client.evict(allocate_size)
            self._size_limit = allocate_size
        return self._size_limit

    def _new_object_id(self, session_id, data_key):
        """
        Calc unique object id for chunks
        """
        while True:
            new_id = plasma.ObjectID.from_random()
            if not self._plasma_client.contains(new_id):
                break
        self._mapper_ref.put(session_id, data_key, new_id)
        return new_id

    def _get_object_id(self, session_id, data_key):
        obj_id = self._mapper_ref.get(session_id, data_key)
        if obj_id is None:
            raise KeyError((session_id, data_key))
        return obj_id

    def create(self, session_id, data_key, size):
        obj_id = self._new_object_id(session_id, data_key)
        try:
            self._plasma_client.evict(size)
            buffer = self._plasma_client.create(obj_id, size)
            return buffer
        except plasma_errors.PlasmaStoreFull:
            exc_type = plasma_errors.PlasmaStoreFull
            self._mapper_ref.delete(session_id, data_key)
            logger.warning('Data %s(%d) failed to store to plasma due to StorageFull',
                           data_key, size)
        except:  # noqa: E722
            self._mapper_ref.delete(session_id, data_key)
            raise

        if exc_type is plasma_errors.PlasmaStoreFull:
            raise StorageFull(request_size=size, capacity=self._size_limit, affected_keys=[data_key])

    def seal(self, session_id, data_key):
        obj_id = self._get_object_id(session_id, data_key)
        try:
            self._plasma_client.seal(obj_id)
        except plasma_errors.PlasmaObjectNotFound:
            self._mapper_ref.delete(session_id, data_key)
            raise KeyError((session_id, data_key))

    def get(self, session_id, data_key):
        """
        Get deserialized Mars object from plasma store
        """
        obj_id = self._get_object_id(session_id, data_key)
        obj = self._plasma_client.get(obj_id, serialization_context=self._serialize_context, timeout_ms=10)
        if obj is plasma.ObjectNotAvailable:
            self._mapper_ref.delete(session_id, data_key)
            raise KeyError((session_id, data_key))
        return obj

    def get_buffer(self, session_id, data_key):
        """
        Get raw buffer from plasma store
        """
        obj_id = self._get_object_id(session_id, data_key)
        [buf] = self._plasma_client.get_buffers([obj_id], timeout_ms=10)
        if buf is None:
            self._mapper_ref.delete(session_id, data_key)
            raise KeyError((session_id, data_key))
        return buf

    def get_actual_size(self, session_id, data_key):
        """
        Get actual size of Mars object from plasma store
        """
        buf = None
        try:
            obj_id = self._get_object_id(session_id, data_key)
            [buf] = self._plasma_client.get_buffers([obj_id], timeout_ms=10)
            if buf is None:
                self._mapper_ref.delete(session_id, data_key)
                raise KeyError((session_id, data_key))
            return buf.size
        finally:
            del buf

    def put(self, session_id, data_key, value):
        """
        Put a Mars object into plasma store
        :param session_id: session id
        :param data_key: chunk key
        :param value: Mars object to be put
        """
        data_size = None

        try:
            obj_id = self._new_object_id(session_id, data_key)
        except StorageDataExists:
            obj_id = self._get_object_id(session_id, data_key)
            if self._plasma_client.contains(obj_id):
                logger.debug('Data %s already exists, returning existing', data_key)
                [buffer] = self._plasma_client.get_buffers([obj_id], timeout_ms=10)
                del value
                return buffer
            else:
                logger.warning('Data %s registered but no data found, reconstructed', data_key)
                self._mapper_ref.delete(session_id, data_key)
                obj_id = self._new_object_id(session_id, data_key)

        try:
            try:
                serialized = dataserializer.serialize(value)
            except SerializationCallbackError:
                self._mapper_ref.delete(session_id, data_key)
                raise SerializationFailed(obj=value) from None

            del value
            data_size = serialized.total_bytes
            try:
                buffer = self._plasma_client.create(obj_id, serialized.total_bytes)
                stream = pyarrow.FixedSizeBufferWriter(buffer)
                stream.set_memcopy_threads(6)
                self._pool.submit(serialized.write_to, stream).result()
                self._plasma_client.seal(obj_id)
            finally:
                del serialized
            return buffer
        except plasma_errors.PlasmaStoreFull:
            self._mapper_ref.delete(session_id, data_key)
            logger.warning('Data %s(%d) failed to store to plasma due to StorageFull',
                           data_key, data_size)
            exc = plasma_errors.PlasmaStoreFull
        except:  # noqa: E722
            self._mapper_ref.delete(session_id, data_key)
            raise

        if exc is plasma_errors.PlasmaStoreFull:
            raise StorageFull(request_size=data_size, capacity=self._size_limit,
                              affected_keys=[data_key])

    def contains(self, session_id, data_key):
        """
        Check if given chunk key exists in current plasma store
        """
        try:
            obj_id = self._get_object_id(session_id, data_key)
            if self._plasma_client.contains(obj_id):
                return True
            else:
                self._mapper_ref.delete(session_id, data_key)
                return False
        except KeyError:
            return False

    def delete(self, session_id, data_key):
        self._mapper_ref.delete(session_id, data_key)

    def batch_delete(self, session_id, data_keys):
        self._mapper_ref.batch_delete(session_id, data_keys)



# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ... import __version__ as mars_version
from ...utils import parse_readable_size

DEFAULT_IMAGE = 'marsproject/mars:v' + mars_version


def _remove_nones(cfg):
    return dict((k, v) for k, v in cfg.items() if v is not None)


class RoleConfig:
    """
    Configuration builder for Kubernetes RBAC roles
    """
    def __init__(self, name, namespace, api_groups, resources, verbs):
        self._name = name
        self._namespace = namespace
        self._api_groups = api_groups.split(',')
        self._resources = resources.split(',')
        self._verbs = verbs.split(',')

    def build(self):
        return {
            'kind': 'Role',
            'metadata': {'name': self._name, 'namespace': self._namespace},
            'rules': [{
                'apiGroups': self._api_groups,
                'resources': self._resources,
                'verbs': self._verbs,
            }]
        }


class RoleBindingConfig:
    """
    Configuration builder for Kubernetes RBAC role bindings
    """
    def __init__(self, name, namespace, role_name, service_account_name):
        self._name = name
        self._namespace = namespace
        self._role_name = role_name
        self._service_account_name = service_account_name

    def build(self):
        return {
            'kind': 'RoleBinding',
            'metadata': {'name': self._name, 'namespace': self._namespace},
            'roleRef': {
                'apiGroup': 'rbac.authorization.k8s.io',
                'kind': 'Role',
                'name': self._role_name,
            },
            'subjects': [{
                'kind': 'ServiceAccount',
                'name': self._service_account_name,
                'namespace': self._namespace,
            }]
        }


class NamespaceConfig:
    """
    Configuration builder for Kubernetes namespaces
    """
    def __init__(self, name):
        self._name = name

    def build(self):
        return {
            'kind': 'Namespace',
            'metadata': {
                'name': self._name,
                'labels': {
                    'name': self._name,
                }
            }
        }


class ServiceConfig:
    """
    Configuration builder for Kubernetes services
    """
    def __init__(self, name, service_type, selector, port, target_port=None,
                 protocol=None):
        self._name = name
        self._type = service_type
        self._protocol = protocol or 'TCP'
        self._selector = selector
        self._port = port
        self._target_port = target_port

    def build(self):
        return {
            'kind': 'Service',
            'metadata': {
                'name': self._name,
            },
            'spec': _remove_nones({
                'type': self._type,
                'selector': self._selector,
                'ports': [
                    _remove_nones({
                        'protocol': self._protocol,
                        'port': self._port,
                        'targetPort': self._target_port,
                    }),
                ]
            }),
        }


class ResourceConfig:
    """
    Configuration builder for Kubernetes computation resources
    """
    def __init__(self, cpu, memory):
        self._cpu = cpu
        self._memory, ratio = parse_readable_size(memory) if memory is not None else (None, False)
        assert not ratio

    def build(self):
        return {
            'cpu': f'{int(self._cpu * 1000)}m',
            'memory': str(int(self._memory)),
        }


class PortConfig:
    """
    Configuration builder for Kubernetes ports definition for containers
    """
    def __init__(self, container_port):
        self._container_port = int(container_port)

    def build(self):
        return {
            'containerPort': self._container_port,
        }


class VolumeConfig:
    """
    Base configuration builder for Kubernetes volumes
    """
    def __init__(self, name, mount_path):
        self.name = name
        self.mount_path = mount_path

    def build(self):
        raise NotImplementedError

    def build_mount(self):
        return {
            'name': self.name,
            'mountPath': self.mount_path,
        }


class HostPathVolumeConfig(VolumeConfig):
    """
    Configuration builder for Kubernetes host volumes
    """
    def __init__(self, name, mount_path, host_path, volume_type=None):
        super().__init__(name, mount_path)
        self._host_path = host_path
        self._volume_type = volume_type or 'DirectoryOrCreate'

    def build(self):
        return {
            'name': self.name,
            'hostPath': {'path': self._host_path, 'type': self._volume_type},
        }


class EmptyDirVolumeConfig(VolumeConfig):
    """
    Configuration builder for Kubernetes empty-dir volumes
    """
    def __init__(self, name, mount_path, use_memory=False):
        super().__init__(name, mount_path)
        self._medium = 'Memory' if use_memory else None

    def build(self):
        result = {
            'name': self.name,
            'emptyDir': {}
        }
        if self._medium:
            result['emptyDir']['medium'] = self._medium
        return result


class ContainerEnvConfig:
    """
    Configuration builder for Kubernetes container environments
    """
    def __init__(self, name, value=None, field_path=None):
        self._name = name
        self._value = value
        self._field_path = field_path

    def build(self):
        result = dict(name=self._name)
        if self._value is not None:
            result['value'] = str(self._value)
        elif self._field_path is not None:  # pragma: no branch
            result['valueFrom'] = {'fieldRef': {'fieldPath': self._field_path}}
        return result


class ProbeConfig:
    """
    Base configuration builder for Kubernetes liveness and readiness probes
    """
    def __init__(self, initial_delay=5, period=1, timeout=None,
                 success_thresh=None, failure_thresh=None):
        self._initial_delay = initial_delay
        self._period = period
        self._timeout = timeout
        self._success_thresh = success_thresh
        self._failure_thresh = failure_thresh

    def build(self):
        return _remove_nones({
            'initialDelaySeconds': self._initial_delay,
            'periodSeconds': self._period,
            'timeoutSeconds': self._timeout,
            'successThreshold': self._success_thresh,
            'failureThreshold': self._failure_thresh,
        })


class ExecProbeConfig(ProbeConfig):
    """
    Configuration builder for Kubernetes probes by executing commands
    """
    def __init__(self, command, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._command = command

    def build(self):
        result = {
            'exec': {'command': self._command}
        }
        result.update(super().build())
        return result


class TcpProbeConfig(ProbeConfig):
    """
    Configuration builder for Kubernetes probes by checking TCP ports
    """
    def __init__(self, port, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._port = port

    def build(self):
        result = {
            'tcpSocket': {'port': self._port}
        }
        result.update(super().build())
        return result


class ReplicationControllerConfig:
    """
    Base configuration builder for Kubernetes replication controllers
    """
    def __init__(self, name, image, replicas, resource_request=None, resource_limit=None,
                 liveness_probe=None, readiness_probe=None, pre_stop_command=None):
        self._name = name
        self._image = image
        self._replicas = replicas
        self._ports = []
        self._volumes = []
        self._envs = dict()

        self.add_default_envs()

        self._resource_request = resource_request
        self._resource_limit = resource_limit

        self._liveness_probe = liveness_probe
        self._readiness_probe = readiness_probe

        self._pre_stop_command = pre_stop_command

    def add_env(self, name, value=None, field_path=None):
        self._envs[name] = ContainerEnvConfig(name, value=value, field_path=field_path)

    def add_simple_envs(self, envs):
        for k, v in envs.items() or ():
            self.add_env(k, v)

    def add_port(self, container_port):
        self._ports.append(PortConfig(container_port))

    def add_default_envs(self):
        pass  # pragma: no cover

    def add_volume(self, vol):
        self._volumes.append(vol)

    def build_container_command(self):
        raise NotImplementedError

    def build_container(self):
        resources_dict = {
            'requests': self._resource_request.build() if self._resource_request else None,
            'limits': self._resource_limit.build() if self._resource_limit else None,
        }
        lifecycle_dict = _remove_nones({
            'preStop': {
                'exec': {'command': self._pre_stop_command},
            } if self._pre_stop_command else None,
        })
        return _remove_nones({
            'command': self.build_container_command(),
            'env': [env.build() for env in self._envs.values()] or None,
            'image': self._image,
            'name': self._name,
            'resources': dict((k, v) for k, v in resources_dict.items() if v) or None,
            'ports': [p.build() for p in self._ports] or None,
            'volumeMounts': [vol.build_mount() for vol in self._volumes] or None,
            'livenessProbe': self._liveness_probe.build() if self._liveness_probe else None,
            'readinessProbe': self._readiness_probe.build() if self._readiness_probe else None,
            'lifecycle': lifecycle_dict or None,
        })

    def build_template_spec(self):
        result = {
            'containers': [self.build_container()],
            'volumes': [vol.build() for vol in self._volumes]
        }
        return dict((k, v) for k, v in result.items() if v)

    def build(self):
        return {
            'kind': 'ReplicationController',
            'metadata': {
                'name': self._name,
            },
            'spec': {
                'replicas': int(self._replicas),
                'selector': {'name': self._name},
                'template': {
                    'metadata': {
                        'labels': {'name': self._name},
                    },
                    'spec': self.build_template_spec()
                }
            },
        }


class MarsReplicationControllerConfig(ReplicationControllerConfig):
    """
    Base configuration builder for replication controllers for Mars
    """
    rc_name = None

    def __init__(self, replicas, cpu=None, memory=None, limit_resources=False,
                 image=None, modules=None, volumes=None, service_port=None, **kwargs):
        self._cpu = cpu
        self._memory, ratio = parse_readable_size(memory) if memory is not None else (None, False)
        assert not ratio

        if isinstance(modules, str):
            self._modules = modules.split(',')
        else:
            self._modules = modules

        res = ResourceConfig(cpu, memory) if cpu or memory else None

        self._service_port = service_port

        super().__init__(
            self.rc_name, image or DEFAULT_IMAGE, replicas,
            resource_request=res, resource_limit=res if limit_resources else None,
            readiness_probe=self.config_readiness_probe(), **kwargs
        )
        if service_port:
            self.add_port(service_port)

        for vol in volumes or ():
            self.add_volume(vol)

    def add_default_envs(self):
        self.add_env('MARS_K8S_POD_NAME', field_path='metadata.name')
        self.add_env('MARS_K8S_POD_NAMESPACE', field_path='metadata.namespace')
        self.add_env('MARS_K8S_POD_IP', field_path='status.podIP')

        if self._service_port:
            self.add_env('MARS_K8S_SERVICE_PORT', str(self._service_port))

        self.add_env('MARS_CONTAINER_IP', field_path='status.podIP')

        if self._cpu:
            self.add_env('MKL_NUM_THREADS', str(self._cpu))
            self.add_env('MARS_CPU_TOTAL', str(self._cpu))
            if getattr(self, 'stat_type', 'cgroup') == 'cgroup':
                self.add_env('MARS_USE_CGROUP_STAT', '1')

        if self._memory:
            self.add_env('MARS_MEMORY_TOTAL', str(int(self._memory)))

        if self._modules:
            self.add_env('MARS_LOAD_MODULES', ','.join(self._modules))

    def config_readiness_probe(self):
        raise NotImplementedError

    @staticmethod
    def get_local_app_module(mod_name):
        return __name__.rsplit('.', 1)[0] + '.' + mod_name


class MarsSchedulersConfig(MarsReplicationControllerConfig):
    """
    Configuration builder for Mars scheduler service
    """
    rc_name = 'marsscheduler'

    def config_readiness_probe(self):
        readiness_cmd = [
            '/srv/entrypoint.sh', self.get_local_app_module('probe'),
        ]
        return ExecProbeConfig(readiness_cmd, timeout=5, failure_thresh=3)

    def build_container_command(self):
        cmd = [
            '/srv/entrypoint.sh', self.get_local_app_module('scheduler'),
        ]
        if self._service_port:
            cmd += ['-p', str(self._service_port)]
        return cmd


class MarsWorkersConfig(MarsReplicationControllerConfig):
    """
    Configuration builder for Mars worker service
    """
    rc_name = 'marsworker'

    def __init__(self, *args, **kwargs):
        spill_volumes = kwargs.pop('spill_volumes', None) or ()
        mount_shm = kwargs.pop('mount_shm', True)
        self._limit_resources = kwargs['limit_resources'] = kwargs.get('limit_resources', True)
        worker_cache_mem = kwargs.pop('worker_cache_mem', None)

        super().__init__(*args, **kwargs)

        self._spill_volumes = []
        for idx, vol in enumerate(spill_volumes):
            if isinstance(vol, str):
                path = f'/mnt/hostpath{idx}'
                self.add_volume(HostPathVolumeConfig(f'host-path-vol-{idx}', path, vol))
                self._spill_volumes.append(path)
            else:
                self.add_volume(vol)
                self._spill_volumes.append(vol.mount_path)
        if self._spill_volumes:
            self.add_env('MARS_SPILL_DIRS', ':'.join(self._spill_volumes))

        if mount_shm:
            self.add_volume(EmptyDirVolumeConfig('shm-volume', '/dev/shm', True))
        if worker_cache_mem:
            self.add_env('MARS_CACHE_MEM_SIZE', worker_cache_mem)

    def config_readiness_probe(self):
        readiness_cmd = [
            '/srv/entrypoint.sh', self.get_local_app_module('probe'),
        ]
        return ExecProbeConfig(readiness_cmd, timeout=60, failure_thresh=3)

    def build_container_command(self):
        cmd = [
            '/srv/entrypoint.sh', self.get_local_app_module('worker'),
        ]
        if self._service_port:
            cmd += ['-p', str(self._service_port)]
        return cmd


class MarsWebsConfig(MarsReplicationControllerConfig):
    """
    Configuration builder for Mars web service
    """
    rc_name = 'marsweb'

    def config_readiness_probe(self):
        if self._service_port:
            return TcpProbeConfig(self._service_port, timeout=60, failure_thresh=3)

    def build_container_command(self):
        cmd = [
            '/srv/entrypoint.sh', self.get_local_app_module('web'),
        ]
        if self._service_port:
            cmd += ['-p', str(self._service_port)]
        return cmd



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools

import numpy as np

from ....tensor.operands import TensorOperand, TensorOperandMixin
from ....tensor import tensor as astensor
from ....tiles import TilesError
from ....utils import check_chunks_unknown_shape
from ...utils import check_array


class PairwiseDistances(TensorOperand, TensorOperandMixin):
    _op_module_ = 'learn'

    @staticmethod
    def _return_float_dtype(X, Y):
        """
        1. If dtype of X and Y is float32, then dtype float32 is returned.
        2. Else dtype float is returned.
        """

        X = astensor(X)

        if Y is None:
            Y_dtype = X.dtype
        else:
            Y = astensor(Y)
            Y_dtype = Y.dtype

        if X.dtype == Y_dtype == np.float32:
            dtype = np.float32
        else:
            dtype = np.float

        return X, Y, dtype

    @staticmethod
    def check_pairwise_arrays(X, Y, precomputed=False, dtype=None):
        X, Y, dtype_float = PairwiseDistances._return_float_dtype(X, Y)

        estimator = 'check_pairwise_arrays'
        if dtype is None:
            dtype = dtype_float

        if Y is X or Y is None:
            X = Y = check_array(X, accept_sparse=True, dtype=dtype,
                                estimator=estimator)
        else:
            X = check_array(X, accept_sparse=True, dtype=dtype,
                            estimator=estimator)
            Y = check_array(Y, accept_sparse=True, dtype=dtype,
                            estimator=estimator)

        if precomputed:
            if X.shape[1] != Y.shape[0]:
                raise ValueError("Precomputed metric requires shape "
                                 f"(n_queries, n_indexed). Got ({X.shape[0]}, {X.shape[1]}) "
                                 f"for {Y.shape[0]} indexed.")
        elif X.shape[1] != Y.shape[1]:
            raise ValueError("Incompatible dimension for X and Y matrices: "
                             f"X.shape[1] == {X.shape[1]} while Y.shape[1] == {Y.shape[1]}")

        return X, Y

    @classmethod
    def _tile_one_chunk(cls, op):
        out = op.outputs[0]
        chunk_op = op.copy().reset_key()
        chunk = chunk_op.new_chunk([op.x.chunks[0], op.y.chunks[0]],
                                   shape=out.shape, order=out.order,
                                   index=(0, 0))
        new_op = op.copy()
        return new_op.new_tensors(op.inputs, shape=out.shape,
                                  order=out.order, chunks=[chunk],
                                  nsplits=tuple((s,) for s in out.shape))

    @classmethod
    def _tile_chunks(cls, op, x, y):
        out = op.outputs[0]
        out_chunks = []
        for idx in itertools.product(range(x.chunk_shape[0]),
                                     range(y.chunk_shape[0])):
            xi, yi = idx

            chunk_op = op.copy().reset_key()
            chunk_inputs = [x.cix[xi, 0], y.cix[yi, 0]]
            out_chunk = chunk_op.new_chunk(
                chunk_inputs, shape=(chunk_inputs[0].shape[0],
                                     chunk_inputs[1].shape[0],),
                order=out.order, index=idx)
            out_chunks.append(out_chunk)

        new_op = op.copy()
        return new_op.new_tensors(op.inputs, shape=out.shape,
                                  order=out.order, chunks=out_chunks,
                                  nsplits=(x.nsplits[0], y.nsplits[0]))

    @classmethod
    def _rechunk_cols_into_one(cls, x, y):
        y_is_x = y is x
        if x.chunk_shape[1] != 1 or y.chunk_shape[1] != 1:
            check_chunks_unknown_shape([x, y], TilesError)

            x = x.rechunk({1: x.shape[1]})._inplace_tile()
            if y_is_x:
                y = x
            else:
                y = y.rechunk({1: y.shape[1]})._inplace_tile()

        return x, y



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from pandas.api.types import is_list_like

from ... import opcodes as OperandDef
from ...core import OutputType
from ...serialize import KeyField, AnyField
from ...tensor.core import TENSOR_TYPE
from ...tiles import TilesError
from ...utils import check_chunks_unknown_shape
from ..core import SERIES_TYPE, INDEX_TYPE
from ..operands import DataFrameOperand, DataFrameOperandMixin


class DataFrameIsin(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = OperandDef.ISIN

    _input = KeyField('input')
    _values = AnyField('values')

    def __init__(self, values=None, output_types=None, **kw):
        if output_types is None:
            output_types = [OutputType.series]
        super().__init__(_values=values, _output_types=output_types, **kw)

    @property
    def input(self):
        return self._input

    @property
    def values(self):
        return self._values

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._input = self._inputs[0]
        if len(inputs) == 2:
            self._values = self._inputs[1]

    def __call__(self, elements):
        inputs = [elements]
        if isinstance(self._values, (SERIES_TYPE, TENSOR_TYPE, INDEX_TYPE)):
            inputs.append(self._values)
        return self.new_series(inputs, shape=elements.shape, dtype=np.dtype('bool'),
                               index_value=elements.index_value, name=elements.name)

    @classmethod
    def tile(cls, op):
        in_elements = op.input
        out_elements = op.outputs[0]

        values = op.values
        if len(op.inputs) == 2:
            # make sure arg has known shape when it's a md.Series
            check_chunks_unknown_shape([op.values], TilesError)
            values = op.values.rechunk(op.values.shape)._inplace_tile()

        out_chunks = []
        for chunk in in_elements.chunks:
            chunk_op = op.copy().reset_key()
            chunk_inputs = [chunk]
            if len(op.inputs) == 2:
                chunk_inputs.append(values.chunks[0])
            out_chunk = chunk_op.new_chunk(chunk_inputs, shape=chunk.shape,
                                           dtype=out_elements.dtype,
                                           index_value=chunk.index_value,
                                           name=out_elements.name,
                                           index=chunk.index)
            out_chunks.append(out_chunk)

        new_op = op.copy()
        return new_op.new_seriess(op.inputs, out_elements.shape,
                                  nsplits=in_elements.nsplits,
                                  chunks=out_chunks, dtype=out_elements.dtype,
                                  index_value=out_elements.index_value, name=out_elements.name)

    @classmethod
    def execute(cls, ctx, op):
        elements = ctx[op.inputs[0].key]
        if len(op.inputs) == 2:
            values = ctx[op.inputs[1].key]
        else:
            values = op.values
        try:
            ctx[op.outputs[0].key] = elements.isin(values)
        except ValueError:
            # buffer read-only
            ctx[op.outputs[0].key] = elements.copy().isin(values.copy())


def isin(elements, values):
    # TODO(hetao): support more type combinations, for example, DataFrame.isin.
    if is_list_like(values):
        values = list(values)
    elif not isinstance(values, (SERIES_TYPE, TENSOR_TYPE, INDEX_TYPE)):
        raise TypeError('only list-like objects are allowed to be passed to isin(), '
                        f'you passed a [{type(values)}]')
    if not isinstance(elements, SERIES_TYPE):  # pragma: no cover
        raise NotImplementedError(f'Unsupported parameter types: {type(elements)} and {type(values)}')
    op = DataFrameIsin(values)
    return op(elements)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Iterable

from pandas.core import groupby as pd_groupby

from ... import opcodes
from ...core import OutputType
from ...serialize import AnyField
from ..operands import DataFrameOperandMixin, DataFrameOperand
from ..utils import parse_index


class GroupByIndex(DataFrameOperandMixin, DataFrameOperand):
    _op_type_ = opcodes.INDEX
    _op_module_ = 'dataframe.groupby'

    _selection = AnyField('selection')

    def __init__(self, selection=None, output_types=None, **kw):
        super().__init__(_selection=selection, _output_types=output_types, **kw)

    @property
    def selection(self):
        return self._selection

    @property
    def groupby_params(self):
        params = self.inputs[0].op.groupby_params
        params['selection'] = self.selection
        return params

    def build_mock_groupby(self, **kwargs):
        groupby_op = self.inputs[0].op
        return groupby_op.build_mock_groupby(**kwargs)[self.selection]

    def __call__(self, groupby):
        indexed = groupby.op.build_mock_groupby()[self.selection]

        if isinstance(indexed, pd_groupby.SeriesGroupBy):
            self.output_types = [OutputType.series_groupby]
            params = dict(shape=(groupby.shape[0],), name=self.selection,
                          dtype=groupby.dtypes[self.selection],
                          index_value=groupby.index_value,
                          key_dtypes=groupby.key_dtypes)
        else:
            self.output_types = [OutputType.dataframe_groupby]

            if isinstance(self.selection, Iterable) and not isinstance(self.selection, str):
                item_list = list(self.selection)
            else:
                item_list = [self.selection]

            params = groupby.params.copy()
            params['dtypes'] = new_dtypes = groupby.dtypes[item_list]
            params['selection'] = self.selection
            params['shape'] = (groupby.shape[0], len(item_list))
            params['columns_value'] = parse_index(new_dtypes.index, store_data=True)

        return self.new_tileable([groupby], **params)

    @classmethod
    def tile(cls, op):
        in_groupby = op.inputs[0]
        out_groupby = op.outputs[0]

        chunks = []
        for c in in_groupby.chunks:
            if op.output_types[0] == OutputType.series_groupby:
                params = dict(shape=(c.shape[0],), name=op.selection,
                              index=(c.index[0],), dtype=c.dtypes[op.selection],
                              index_value=c.index_value, key_dtypes=c.key_dtypes)
            else:
                params = c.params.copy()
                params['dtypes'] = out_groupby.dtypes
                params['selection'] = op.selection
                params['shape'] = (c.shape[0], len(op.selection))

            new_op = op.copy().reset_key()
            chunks.append(new_op.new_chunk([c], **params))

        new_op = op.copy().reset_key()
        params = out_groupby.params.copy()
        new_nsplits = (in_groupby.nsplits[0], (len(op.selection),)) if out_groupby.ndim == 2 \
            else (in_groupby.nsplits[0],)
        params.update(dict(chunks=chunks, nsplits=new_nsplits))
        return new_op.new_tileables([in_groupby], **params)

    @classmethod
    def execute(cls, ctx, op):
        in_data = ctx[op.inputs[0].key]
        ctx[op.outputs[0].key] = in_data[op.selection]


def df_groupby_getitem(df_groupby, item):
    try:
        hash(item)
        hashable = True
    except TypeError:
        hashable = False

    if hashable and item in df_groupby.dtypes:
        output_types = [OutputType.series_groupby]
    elif isinstance(item, Iterable) and all(it in df_groupby.dtypes for it in item):
        output_types = [OutputType.dataframe_groupby]
    else:
        raise NameError(f'Cannot slice groupby with {item!r}')

    if df_groupby.selection:
        raise IndexError(f'Column(s) {df_groupby.selection!r} already selected')

    op = GroupByIndex(selection=item, output_types=output_types)
    return op(df_groupby)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pickle
import sys
import weakref
from collections import OrderedDict
from datetime import tzinfo
from io import BytesIO

import numpy as np
cimport numpy as np
cimport cython
import cloudpickle
import pandas as pd
from pandas.api.extensions import ExtensionDtype
from pandas.arrays import IntervalArray

from ..utils import serialize_function
from .core cimport ProviderType, ValueType, Identity, List, Tuple, Dict, \
    Reference, KeyPlaceholder, AttrWrapper, Provider, Field, OneOfField, \
    ReferenceField, IdentityField, ListField, TupleField, \
    get_serializable_by_index
from .core import HasKey, HasData
from .protos.value_pb2 import Value
from .dataserializer import dumps as datadumps, loads as dataloads

try:
    from pandas.tseries.offsets import Tick as PDTick
    from pandas.tseries.frequencies import to_offset
except ImportError:
    PDTick = to_offset = None


cdef dict PRIMITIVE_TYPE_TO_VALUE_FIELD = {
    ValueType.bool: 'b',
    ValueType.int8: 'i',
    ValueType.int16: 'i',
    ValueType.int32: 'i',
    ValueType.int64: 'i',
    ValueType.uint8: 'i',
    ValueType.uint16: 'i',
    ValueType.uint32: 'i',
    ValueType.uint64: 'i',
    ValueType.float16: 'f',
    ValueType.float32: 'f',
    ValueType.float64: 'f',
    ValueType.bytes: 's',
    ValueType.unicode: 'u',
}


cdef class ProtobufSerializeProvider(Provider):
    def __init__(self, data_serial_type=None, pickle_protocol=None):
        self.type = ProviderType.protobuf
        self.data_serial_type = data_serial_type
        self.pickle_protocol = pickle_protocol

    cdef inline void _set_slice(self, slice value, obj, tp=None):
        if value.start is not None:
            if isinstance(value.start, (int, np.integer)):
                obj.slice.start_val = value.start
            elif isinstance(value.start, str):
                obj.slice.start_label = value.start
            else:
                # dump if not integer or string
                obj.slice.start_obj = pickle.dumps(value.start, protocol=self.pickle_protocol)
        if value.stop is not None:
            if isinstance(value.stop, (int, np.integer)):
                obj.slice.stop_val = value.stop
            elif isinstance(value.stop, str):
                obj.slice.stop_label = value.stop
            else:
                # dump if not integer or string
                obj.slice.stop_obj = pickle.dumps(value.stop, protocol=self.pickle_protocol)
        if value.step is not None:
            obj.slice.step_val = value.step
        if all(it is None for it in (value.start, value.stop, value.step)):
            obj.slice.is_null = True

    cdef inline slice _get_slice(self, obj):
        if obj.slice.is_null:
            return slice(None)

        start_key = obj.slice.WhichOneof('start')
        if start_key is not None:
            start = getattr(obj.slice, start_key)
            if start_key == 'start_obj':
                start = pickle.loads(start)
        else:
            start = None
        stop_key = obj.slice.WhichOneof('stop')
        if stop_key is not None:
            stop = getattr(obj.slice, stop_key)
            if stop_key == 'stop_obj':
                stop = pickle.loads(stop)
        else:
            stop = None
        step = obj.slice.step_val if obj.slice.WhichOneof('step') else None
        return slice(start, stop, step)

    cdef inline void _set_arr(self, np.ndarray value, obj, tp=None):
        # special case for np.unicode and np.bytes_
        # cuz datadumps may fail due to pyarrow
        if value.ndim == 0 and value.dtype.kind in ('U', 'S'):
            value = value.astype(object)
        obj.arr = datadumps(value, serial_type=self.data_serial_type,
                            pickle_protocol=self.pickle_protocol)

    cdef inline np.ndarray _get_arr(self, obj):
        cdef object x

        x = obj.arr
        return dataloads(x) if x is not None and len(x) > 0 else None

    cdef inline void _set_dtype(self, value, obj, tp=None):
        if not isinstance(value, ExtensionDtype) and 'V' not in value.str:
            dtype = value.str
            if isinstance(dtype, unicode):
                dtype = dtype.encode('utf-8')
            obj.dtype = dtype
        else:
            obj.dtype = pickle.dumps(value, protocol=self.pickle_protocol)

    cdef inline object _get_dtype(self, obj):
        if obj.dtype is None or len(obj.dtype) == 0:
            return
        try:
            return np.dtype(obj.dtype)
        except (TypeError, ValueError):
            return pickle.loads(obj.dtype)

    cdef inline void _set_index(self, value, obj, tp=None) except *:
        obj.index = datadumps(value, serial_type=self.data_serial_type,
                              pickle_protocol=self.pickle_protocol)

    cdef inline _get_index(self, obj):
        x = obj.index
        return dataloads(x) if x is not None and len(x) > 0 else None

    cdef inline void _set_series(self, value, obj, tp=None) except *:
        obj.series = datadumps(value, serial_type=self.data_serial_type,
                               pickle_protocol=self.pickle_protocol)

    cdef inline _get_series(self, obj):
        x = obj.series
        return dataloads(x) if x is not None and len(x) > 0 else None

    cdef inline void _set_dataframe(self, value, obj, tp=None) except *:
        obj.dataframe = datadumps(value, serial_type=self.data_serial_type,
                                  pickle_protocol=self.pickle_protocol)

    cdef inline object _get_dataframe(self, obj):
        x = obj.dataframe
        return dataloads(x) if x is not None and len(x) > 0 else None

    cdef inline void _set_key(self, object value, obj, tp=None) except *:
        obj.key.key = value.key
        obj.key.id = value.id

    cdef inline KeyPlaceholder _get_key(self, obj):
        if obj.key.key is None or len(obj.key.key) == 0:
            return
        return KeyPlaceholder(obj.key.key, obj.key.id)

    cdef inline void _set_datetime64(self, value, obj, tp=None):
        if isinstance(value, pd.Timestamp):
            # convert to np.datetime64
            value = value.to_datetime64()
        bio = BytesIO()
        np.save(bio, value)
        obj.datetime64 = bio.getvalue()

    cdef inline object _get_datetime64(self, obj):
        cdef object x

        x = obj.datetime64
        if x is not None and len(x) > 0:
            # np.load will return a ndarray
            value = np.load(BytesIO(x))
            return value.dtype.type(value)
        else:
            return None

    cdef inline void _set_timedelta64(self, value, obj, tp=None):
        if isinstance(value, pd.Timedelta):
            # convert to np.timedelta64
            value = value.to_timedelta64()
        bio = BytesIO()
        np.save(bio, value)
        obj.timedelta64 = bio.getvalue()

    cdef inline object _get_timedelta64(self, obj):
        cdef object x

        x = obj.timedelta64
        if x is not None and len(x) > 0:
            # np.load will return a ndarray
            value = np.load(BytesIO(x))
            return value.dtype.type(value)
        else:
            return None

    cdef inline void _set_complex(self, value, obj, tp=None):
        obj.c.real = value.real
        obj.c.imag = value.imag

    cdef inline object _get_complex(self, obj):
        return complex(obj.c.real, obj.c.imag)

    cdef inline void _set_function(self, value, obj, tp=None):
        obj.function = serialize_function(value, pickle_protocol=self.pickle_protocol)

    cdef inline object _get_function(self, obj):
        x = obj.function
        return cloudpickle.loads(x) if x is not None and len(x) > 0 else None

    cdef inline void _set_tzinfo(self, value, obj, tp=None):
        obj.tzinfo = pickle.dumps(value, protocol=self.pickle_protocol)

    cdef inline object _get_tzinfo(self, obj):
        x = obj.tzinfo
        return pickle.loads(x) if x is not None and len(x) > 0 else None

    cdef inline void _set_interval_arr(self, value, obj, tp=None):
        obj.interval_arr.left = datadumps(value.left, serial_type=self.data_serial_type,
                                          pickle_protocol=self.pickle_protocol)
        obj.interval_arr.right = datadumps(value.right, serial_type=self.data_serial_type,
                                           pickle_protocol=self.pickle_protocol)
        obj.interval_arr.closed = value.closed
        self._set_dtype(value.dtype, obj.interval_arr)

    cdef inline object _get_interval_arr(self, obj):
        interval_arr = obj.interval_arr
        if not interval_arr.left:
            return
        left = dataloads(interval_arr.left)
        right = dataloads(interval_arr.right)
        closed = interval_arr.closed
        dtype = self._get_dtype(interval_arr)
        return IntervalArray.from_arrays(left, right,
                                         closed=closed, dtype=dtype)

    cdef inline void _set_freq(self, value, obj, tp=None):
        obj.freq = value.freqstr

    cdef inline object _get_freq(self, obj):
        return to_offset(obj.freq)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.nonecheck(False)
    cdef inline void _set_list(self, list value, obj, tp=None, bint weak_ref=False) except *:
        cdef list res

        # list is special, cuz the internal object can be a specified type or any type
        if isinstance(obj, Value):
            for val in value:
                if weak_ref:
                    val = val()
                it_obj = obj.list.value.add()
                self._set_value(val, it_obj, tp=tp.type if tp is not None else tp)
        else:
            try:
                obj.extend(value)
            except TypeError:
                # not primitive type
                res = []
                for val in value:
                    if weak_ref:
                        val = val()
                    it_obj = Value()
                    self._set_value(val, it_obj, tp=tp.type if tp is not None else tp)
                    res.append(it_obj)
                obj.extend(res)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.nonecheck(False)
    cdef inline list _get_list(self, obj, tp, list callbacks, bint weak_ref):
        cdef list res
        cdef int i
        cdef object it_obj

        obj = obj.list.value if isinstance(obj, Value) else obj
        res = []
        for i, it_obj in enumerate(obj):
            if not isinstance(it_obj, Value):
                res.append(it_obj if not weak_ref else weakref.ref(it_obj))
                continue

            val = self._get_value(it_obj, tp.type if tp is not None else tp,
                                  callbacks, weak_ref)
            res.append(val)
            if isinstance(val, KeyPlaceholder):
                def cb(j, v):
                    def inner(subs):
                        o = subs[v.key, v.id]
                        if weak_ref:
                            o = weakref.ref(o)
                        res[j] = o
                    return inner
                callbacks.append(cb(i, val))
        return res

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.nonecheck(False)
    cdef inline void _set_tuple(self, tuple value, obj, tp=None, bint weak_ref=False) except *:
        cdef list res

        if isinstance(obj, Value):
            obj.list.is_tuple = True
            if tp is not None and isinstance(tp.type, tuple):
                if len(tp.type) != len(value):
                    raise ValueError(f'Value length should be {len(tp.type)} '
                                     f'according to type definition, got {len(value)}')
                for it_type, val in zip(tp.type, value):
                    if weak_ref:
                        val = val()
                    if it_type == ValueType.key:
                        raise TypeError('`key` type is not allowed in a tuple')
                    it_obj = obj.list.value.add()
                    self._set_typed_value(val, it_obj, it_type)
            else:
                if tp == ValueType.key:
                    raise TypeError('`key` type is not allowed in a tuple')
                for val in value:
                    if weak_ref:
                        val = val()
                    it_obj = obj.list.value.add()
                    self._set_value(val, it_obj, tp=tp.type if tp is not None else tp)
        else:
            try:
                if weak_ref:
                    value = tuple(val() for val in value)
                obj.extend(value)
            except TypeError:
                # not primitive type
                res = []
                for val in value:
                    if weak_ref:
                        val = val()
                    it_obj = Value()
                    self._set_value(val, it_obj, tp=tp.type if tp is not None else tp)
                    res.append(it_obj)
                obj.extend(res)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.nonecheck(False)
    cdef inline tuple _get_tuple(self, obj, tp, list callbacks, bint weak_ref):
        cdef list res

        values = obj.list.value if isinstance(obj, Value) else obj

        res = []
        if tp is not None and isinstance(tp.type, tuple):
            for it_type, it_obj in zip(tp.type, values):
                if isinstance(it_obj, Value):
                    res.append(self._get_value(it_obj, it_type,
                                               callbacks, weak_ref))
                else:
                    if weak_ref:
                        it_obj = weakref.ref(it_obj)
                    res.append(it_obj)
        else:
            for it_obj in values:
                if isinstance(it_obj, Value):
                    res.append(self._get_value(it_obj, tp.type if tp is not None else tp,
                                               callbacks, weak_ref))
                else:
                    if weak_ref:
                        it_obj = weakref.ref(it_obj)
                    res.append(it_obj)

        return tuple(res)

    cdef inline void _set_dict(self, value, obj, tp=None, bint weak_ref=False) except *:
        for k, v in value.items():
            if weak_ref:
                k, v = k(), v()
            key_obj = obj.dict.keys.value.add()
            if tp is not None and tp.key_type == ValueType.key:
                raise TypeError('`key` type is not allowd as dict key')
            self._set_value(k, key_obj, tp=tp.key_type if tp is not None else tp)
            value_obj = obj.dict.values.value.add()
            self._set_value(v, value_obj, tp=tp.value_type if tp is not None else tp)

    cdef inline object _get_dict(self, obj, tp, list callbacks, bint weak_ref):
        d = OrderedDict()
        for k_obj, v_obj in zip(obj.dict.keys.value, obj.dict.values.value):
            k = self._get_value(k_obj, tp.key_type if tp is not None else tp,
                                callbacks, weak_ref)
            v = self._get_value(v_obj, tp.value_type if tp is not None else tp,
                                callbacks, weak_ref)
            d[k] = v
            if isinstance(v, KeyPlaceholder):
                def cb(key, value):
                    def inner(subs):
                        o = subs[value.key, value.id]
                        if weak_ref:
                            o = weakref.ref(o)
                        d[key] = o
                    return inner
                callbacks.append(cb(k, v))

        return d

    cdef inline void _set_typed_value(self, value, obj, tp, bint weak_ref=False) except *:
        cdef str value_field

        if type(tp) not in (List, Tuple, Dict) and weak_ref:
            # not iterable, and is weak ref
            value = value()

        if value is None:
            # handle None
            obj.is_null = True
            return

        if tp in PRIMITIVE_TYPE_TO_VALUE_FIELD:
            value_field = PRIMITIVE_TYPE_TO_VALUE_FIELD[tp]
            setattr(obj, value_field, value)
        elif tp is ValueType.slice:
            self._set_slice(value, obj, tp)
        elif tp is ValueType.arr:
            self._set_arr(<np.ndarray>value, obj, tp)
        elif tp is ValueType.dtype:
            self._set_dtype(<np.dtype>value, obj, tp)
        elif tp is ValueType.index:
            self._set_index(value, obj, tp)
        elif tp is ValueType.series:
            self._set_series(value, obj, tp)
        elif tp is ValueType.dataframe:
            self._set_dataframe(value, obj, tp)
        elif tp is ValueType.key:
            self._set_key(value, obj, tp)
        elif tp is ValueType.datetime64:
            self._set_datetime64(value, obj, tp)
        elif tp is ValueType.timedelta64:
            self._set_timedelta64(value, obj, tp)
        elif tp is ValueType.function:
            self._set_function(value, obj, tp)
        elif tp is ValueType.tzinfo:
            self._set_tzinfo(value, obj, tp)
        elif tp is ValueType.interval_arr:
            self._set_interval_arr(value, obj, tp)
        elif tp is ValueType.freq:
            self._set_freq(value, obj, tp)
        elif tp in {ValueType.complex64, ValueType.complex128}:
            self._set_complex(value, obj, tp)
        elif isinstance(tp, Identity):
            value_field = PRIMITIVE_TYPE_TO_VALUE_FIELD[tp.type]
            setattr(obj, value_field, value)
        elif isinstance(tp, List):
            # list type
            self._set_list(<list>value, obj, tp, weak_ref=weak_ref)
        elif isinstance(tp, Tuple):
            # tuple type
            self._set_tuple(<tuple>value, obj, tp, weak_ref=weak_ref)
        elif isinstance(tp, Dict):
            # dict type
            self._set_dict(<dict>value, obj, tp, weak_ref=weak_ref)
        else:
            raise TypeError(f'Unknown type to serialize: {tp}')

    cdef inline void _set_untyped_value(self, value, obj, bint weak_ref=False) except *:
        if not isinstance(value, (list, tuple, dict)) and weak_ref:
            # not iterable, and is weak ref
            value = value()

        # we are not aware of the type, so try to infer it
        if value is None:
            # handle None
            obj.is_null = True
            return

        if isinstance(value, bool):
            obj.b = value
        elif isinstance(value, bytes):
            obj.s = value
        elif isinstance(value, unicode):
            obj.u = value
        elif isinstance(value, int):
            obj.i = value
        elif isinstance(value, float):
            obj.f = value
        elif isinstance(value, complex):
            self._set_complex(value, obj)
        elif isinstance(value, slice):
            self._set_slice(value, obj)
        elif isinstance(value, np.ndarray):
            self._set_arr(value, obj)
        elif isinstance(value, (np.dtype, ExtensionDtype)):
            self._set_dtype(value, obj)
        elif isinstance(value, pd.Index):
            self._set_index(value, obj)
        elif isinstance(value, pd.Series):
            self._set_series(value, obj)
        elif isinstance(value, pd.DataFrame):
            self._set_dataframe(value, obj)
        elif isinstance(value, HasKey):
            self._set_key(value, obj)
        elif isinstance(value, HasData):
            self._set_key(value.data, obj)
        elif isinstance(value, list):
            self._set_list(value, obj, tp=None, weak_ref=weak_ref)
        elif isinstance(value, tuple):
            self._set_tuple(value, obj, tp=None, weak_ref=weak_ref)
        elif isinstance(value, dict):
            self._set_dict(value, obj, tp=None, weak_ref=weak_ref)
        elif isinstance(value, (np.datetime64, pd.Timestamp)):
            self._set_datetime64(value, obj)
        elif isinstance(value, (np.timedelta64, pd.Timedelta)):
            self._set_timedelta64(value, obj)
        elif isinstance(value, np.number):
            self._set_untyped_value(value.item(), obj)
        elif isinstance(value, tzinfo):
            self._set_tzinfo(value, obj)
        elif isinstance(value, IntervalArray):
            self._set_interval_arr(value, obj)
        elif PDTick is not None and isinstance(value, PDTick):
            self._set_freq(value, obj)
        elif callable(value):
            self._set_function(value, obj)
        else:
            raise TypeError(f'Unknown type to serialize: {type(value)}')

    cdef inline void _set_value(cls, value, obj, tp=None, bint weak_ref=False) except *:
        if tp is None:
            cls._set_untyped_value(value, obj, weak_ref=weak_ref)
        else:
            cls._set_typed_value(value, obj, tp, weak_ref=weak_ref)

    cdef inline void _serial_reference_value(self, tag, model, value, value_pb) except *:
        cdef object field_obj

        if model is None:
            if type(value_pb) is Value and not hasattr(value, 'cls'):
                value_pb.typed.type_id = value.__serializable_index__
                value.serialize(self, value_pb.typed.value)
            else:
                if type(value_pb) is Value:
                    value_pb = value_pb.any_ref
                field_obj = value.cls(self)()
                value.serialize(self, obj=field_obj)
                value_pb.type_id = value.__serializable_index__
                value_pb.value = field_obj.SerializeToString()
        else:
            if isinstance(value, model):
                value.serialize(self, obj=value_pb)
            else:
                raise TypeError(f'Does not match type for reference field {tag}: '
                                f'expect {model}, got {type(value)}')

    cpdef serialize_field(self, Field field, model_instance, obj):
        cdef object value
        cdef object val
        cdef Field f
        cdef str tag
        cdef object field_obj
        cdef object add
        cdef object it_obj
        cdef bint matched
        cdef OneOfField oneoffield

        if isinstance(field, OneOfField):
            oneoffield = <OneOfField> field
            value = getattr(model_instance, oneoffield.attr, None)
            matched = False
            for f in oneoffield.fields:
                if isinstance(value, f.type.model):
                    f.serialize(self, model_instance, obj)
                    matched = True
                    return
            if not matched and value is not None:
                raise ValueError(f'Value {value} cannot match any type for OneOfField `{field.tag_name(self)}`')
            return

        value = getattr(model_instance, field.attr, field.default)
        if field.on_serialize and value is not None:
            value = field.on_serialize(value)
        tag = field.tag_name(self)
        try:
            field_obj = getattr(obj, tag)
        except AttributeError:
            raise AttributeError(f'no attribute {tag} for {model_instance}')

        if value is None:
            return

        if isinstance(field, ReferenceField):
            if field.weak_ref:
                field_obj = field_obj()
            self._serial_reference_value(tag, field.type.model, value, field_obj)
        elif isinstance(field, ListField):
            if type(field.type.type) == Reference:
                add = field_obj.list.value.add if isinstance(field_obj, Value) else field_obj.add
                for val in value:
                    if field.weak_ref:
                        val = val()
                    it_obj = add()
                    if val is not None:
                        self._serial_reference_value(tag, field.type.type.model, val, it_obj)
                    elif isinstance(it_obj, Value):
                        it_obj.is_null = True
            else:
                # repeated primitive type
                if not isinstance(value, list):
                    value = list(value)
                self._set_list(value, field_obj, tp=field.type, weak_ref=field.weak_ref)
        elif isinstance(field_obj, Value):
            try:
                self._set_value(value, field_obj, field.type, weak_ref=field.weak_ref)
            except TypeError:
                exc_info = sys.exc_info()
                raise TypeError(f'Failed to set field `{tag}` for {model_instance} with '
                                f'value {value}, reason: {exc_info[1]}') \
                    .with_traceback(exc_info[2]) from exc_info[1]
        elif isinstance(field, TupleField):
            try:
                self._set_tuple(value, field_obj, tp=field.type, weak_ref=field.weak_ref)
            except TypeError:
                exc_info = sys.exc_info()
                raise TypeError(f'Failed to set field `{tag}` for {model_instance} with '
                                f'value {value}, reason: {exc_info[1]}') \
                    .with_traceback(exc_info[2]) from exc_info[1]
        else:
            try:
                setattr(obj, tag, value)
            except TypeError:
                exc_info = sys.exc_info()
                raise TypeError(f'Failed to set field `{tag}` for {model_instance} with '
                                f'value {value}, reason: {exc_info[1]}') \
                    .with_traceback(exc_info[2]) from exc_info[1]

    cpdef serialize_attribute_as_dict(self, model_instance, obj=None):
        cdef object id_field
        cdef str attr
        cdef object d_obj
        cdef dict fields
        cdef str name
        cdef object field
        cdef object value

        if obj is None:
            obj = model_instance.cls(self)()

        if hasattr(model_instance, '_ID_FIELD') and not isinstance(obj, Value):
            for id_field in model_instance._ID_FIELD:
                id_field.serialize(self, model_instance, obj)

        attr = getattr(model_instance, 'attr_tag', None)
        if attr:
            d_obj = getattr(obj, attr)
        else:
            d_obj = obj

        if isinstance(d_obj, Value):
            fields = model_instance._FIELDS
            for name, field in fields.items():
                value = getattr(model_instance, name, None)
                if value is None:
                    continue
                tag = field.tag_name(self)

                k = d_obj.dict.keys.value.add()
                self._set_value(field.tag_name(self), k, tp=ValueType.string)
                v = d_obj.dict.values.value.add()
                if isinstance(field, ReferenceField):
                    if field.weak_ref:
                        value = value()
                    self._serial_reference_value(tag, field.type.model, value, v)
                elif isinstance(field, ListField) and type(field.type.type) == Reference:
                    for val in value:
                        if field.weak_ref:
                            val = val()
                        it_obj = v.list.value.add()
                        if val is not None:
                            self._serial_reference_value(tag, field.type.type.model, val, it_obj)
                        elif isinstance(it_obj, Value):
                            it_obj.is_null = True
                else:
                    self._set_value(value, v, tp=field.type, weak_ref=field.weak_ref)
        else:
            fields = model_instance._FIELDS
            for name, field in fields.items():
                value = getattr(model_instance, name, None)
                if value is not None:
                    field.serialize(self, model_instance, AttrWrapper(d_obj))

        return obj

    cdef inline object _get_typed_value(self, obj, tp, list callbacks, bint weak_ref):
        cdef str value_field

        if obj.is_null:
            return

        if weak_ref:
            ref = weakref.ref
        else:
            ref = lambda x: x

        if tp in PRIMITIVE_TYPE_TO_VALUE_FIELD:
            value_field = PRIMITIVE_TYPE_TO_VALUE_FIELD[tp]
            return ref(getattr(obj, value_field))
        elif tp in {ValueType.complex64, ValueType.complex128}:
            return ref(self._get_complex(obj))
        elif tp is ValueType.slice:
            return ref(self._get_slice(obj))
        elif tp is ValueType.arr:
            return ref(self._get_arr(obj))
        elif tp is ValueType.dtype:
            return ref(self._get_dtype(obj))
        elif tp is ValueType.index:
            return ref(self._get_index(obj))
        elif tp is ValueType.series:
            return ref(self._get_series(obj))
        elif tp is ValueType.dataframe:
            return ref(self._get_dataframe(obj))
        elif tp is ValueType.key:
            return self._get_key(obj)
        elif tp is ValueType.datetime64:
            return ref(self._get_datetime64(obj))
        elif tp is ValueType.timedelta64:
            return ref(self._get_timedelta64(obj))
        elif tp is ValueType.function:
            return ref(self._get_function(obj))
        elif tp is ValueType.tzinfo:
            return ref(self._get_tzinfo(obj))
        elif tp is ValueType.interval_arr:
            return ref(self._get_interval_arr(obj))
        elif tp is ValueType.freq:
            return ref(self._get_freq(obj))
        elif isinstance(tp, Identity):
            value_field = PRIMITIVE_TYPE_TO_VALUE_FIELD[tp.type]
            return ref(getattr(obj, value_field))
        elif isinstance(tp, List):
            # list type
            return self._get_list(obj, tp, callbacks, weak_ref)
        elif isinstance(tp, Tuple):
            # tuple type
            return self._get_tuple(obj, tp, callbacks, weak_ref)
        elif isinstance(tp, Dict):
            # dict type
            return self._get_dict(obj, tp, callbacks, weak_ref)
        else:
            raise TypeError(f'Unknown type to deserialize: {tp}')

    cdef inline object _get_untyped_value(self, obj, list callbacks, bint weak_ref):
        cdef str field

        if weak_ref:
            ref = weakref.ref
        else:
            ref = lambda x: x

        # we are not aware of the type, so try to infer it
        field = obj.WhichOneof('value')

        if field is None or field == 'is_null':
            return
        elif field in 'bif':
            # primitive type
            return ref(getattr(obj, field))
        elif field == 'c':
            return ref(self._get_complex(obj))
        elif field == 's':
            # bytes
            b = obj.s
            if not isinstance(b, bytes):
                b = b.encode('utf-8')
            return ref(b)
        elif field == 'u':
            # unicode
            u = obj.u
            if not isinstance(u, unicode):
                u = u.decode('utf-8')
            return ref(u)
        elif field == 'list':
            if obj.list.is_tuple:
                return self._get_tuple(obj, None, callbacks, weak_ref)
            else:
                return self._get_list(obj, None, callbacks, weak_ref)
        elif field == 'dict':
            return self._get_dict(obj, None, callbacks, weak_ref)
        elif field == 'slice':
            return ref(self._get_slice(obj))
        elif field == 'arr':
            return ref(self._get_arr(obj))
        elif field == 'dtype':
            return ref(self._get_dtype(obj))
        elif field == 'index':
            return ref(self._get_index(obj))
        elif field == 'series':
            return ref(self._get_series(obj))
        elif field == 'dataframe':
            return ref(self._get_dataframe(obj))
        elif field == 'key':
            return self._get_key(obj)
        elif field == 'datetime64':
            return ref(self._get_datetime64(obj))
        elif field == 'timedelta64':
            return ref(self._get_timedelta64(obj))
        elif field == 'function':
            return ref(self._get_function(obj))
        elif field == 'tzinfo':
            return ref(self._get_tzinfo(obj))
        elif field == 'interval_arr':
            return ref(self._get_interval_arr(obj))
        elif field == 'freq':
            return ref(self._get_freq(obj))
        else:
            raise TypeError('Unknown type to deserialize')

    cdef inline object _get_value(self, obj, tp, list callbacks, bint weak_ref):
        if tp is None:
            return self._get_untyped_value(obj, callbacks, weak_ref)
        else:
            return self._get_typed_value(obj, tp, callbacks, weak_ref)

    cdef inline object _on_deserial(self, Field field, x):
        x = x if x is not None else field.default
        if field.on_deserialize is None:
            return x
        return field.on_deserialize(x)

    cdef object _deserial_reference_value(self, model, value_pb, list callbacks, dict key_to_instance):
        cdef object f_obj

        if model is None:
            if type(value_pb) is Value and value_pb.typed.type_id != 0:
                model = get_serializable_by_index(value_pb.typed.type_id)
                f_obj = value_pb.typed.value
            else:
                if type(value_pb) is Value:
                    value_pb = value_pb.any_ref
                if value_pb.type_id == 0:
                    return None
                model = get_serializable_by_index(value_pb.type_id)
                f_obj = model.cls(self)()
                f_obj.ParseFromString(value_pb.value)
        else:
            f_obj = value_pb
        return model.deserialize(self, f_obj, callbacks, key_to_instance)

    def deserialize_field(self, Field field, model_instance, obj, list callbacks, dict key_to_instance):
        cdef str tag
        cdef str f_tag
        cdef Field f
        cdef object f_obj
        cdef object it_obj
        cdef object field_obj
        cdef object value
        cdef OneOfField oneoffield

        if isinstance(field, OneOfField):
            oneoffield = <OneOfField> field
            tag = oneoffield.tag_name(self)
            f_tag = obj.WhichOneof(tag)
            if f_tag is None:
                return
            f = next(f for f in oneoffield.fields if f.tag_name(self) == f_tag)
            f_obj = getattr(obj, f_tag)
            setattr(model_instance, f.attr,
                    f.type.model.deserialize(self, f_obj, callbacks, key_to_instance))
            return

        tag = field.tag_name(self)
        field_obj = getattr(obj, tag)

        if isinstance(field, ReferenceField):
            value = self._deserial_reference_value(field.type.model, field_obj, callbacks, key_to_instance)
            if value is None:
                return
            setattr(model_instance, field.attr, self._on_deserial(field, value))
        elif isinstance(field, ListField):
            if type(field.type.type) == Reference:
                field_obj = field_obj.list.value if isinstance(field_obj, Value) else field_obj
                value = [self._deserial_reference_value(
                    field.type.type.model, it_obj, callbacks, key_to_instance) for it_obj in field_obj]
                setattr(model_instance, field.attr, self._on_deserial(field, value))
            else:
                value = self._get_list(field_obj, field.type, callbacks, field.weak_ref)
                setattr(model_instance, field.attr,
                        self._on_deserial(field, value))
        elif isinstance(field_obj, Value):
            val = self._get_value(field_obj, field.type, callbacks, field.weak_ref)
            setattr(model_instance, field.attr, self._on_deserial(field, val))
            if isinstance(val, KeyPlaceholder):
                def cb(subs):
                    o = self._on_deserial(field, subs[val.key, val.id])
                    if field.weak_ref:
                        o = weakref.ref(o)
                    setattr(model_instance, field.attr, o)

                callbacks.append(cb)
        elif isinstance(field, TupleField):
            setattr(model_instance, field.attr,
                    self._on_deserial(field, self._get_tuple(field_obj, field.type, callbacks,
                                                             field.weak_ref)))
        else:
            setattr(model_instance, field.attr, self._on_deserial(field, getattr(obj, tag)))

    def deserialize_attribute_as_dict(self, model_cls, obj, list callbacks, dict key_to_instance):
        from ..core import Base

        cdef str attr
        cdef object d_obj
        cdef object kw
        cdef tuple id_fields
        cdef IdentityField id_field
        cdef object model_instance
        cdef dict tag_to_fields
        cdef Field field
        cdef Field f
        cdef OneOfField oneoffield
        cdef Field it_field
        cdef str tag
        cdef object o_tag
        cdef object value

        attr = getattr(model_cls, 'attr_tag', None)
        if attr:
            d_obj = getattr(obj, attr)
        else:
            d_obj = obj

        kw = AttrWrapper(dict())
        if hasattr(model_cls, '_ID_FIELD'):
            # get the id field
            if attr is None:
                if d_obj.is_null:
                    return
                id_fields = model_cls._ID_FIELD
                for id_field in id_fields:
                    tag = id_field.tag_name(self)
                    v = next(v for k, v in zip(d_obj.dict.keys.value, d_obj.dict.values.value)
                             if self._get_value(k, ValueType.string, callbacks, id_field.weak_ref) == tag)
                    setattr(kw, id_field.attr, self._get_value(v, id_field.type,
                                                               callbacks, id_field.weak_ref))
            else:
                for id_field in model_cls._ID_FIELD:
                    id_field.deserialize(self, kw, obj, callbacks, key_to_instance)
        model_instance = model_cls(**kw.asdict())

        tag_to_fields = dict()
        for field in model_instance._FIELDS.values():
            if isinstance(field, OneOfField):
                oneoffield = <OneOfField> field
                for f in oneoffield.fields:
                    tag_to_fields[f.tag_name(self)] = f
            else:
                tag_to_fields[field.tag_name(self)] = field

        if isinstance(d_obj, Value):
            if d_obj.is_null:
                return
            for k, v in zip(d_obj.dict.keys.value, d_obj.dict.values.value):
                tag = self._get_value(k, ValueType.string, callbacks, False)
                it_field = tag_to_fields[tag]
                if isinstance(it_field, ReferenceField):
                    value = self._deserial_reference_value(it_field.type.model, v, callbacks, key_to_instance)
                    if value is None:
                        return
                    setattr(model_instance, it_field.attr, self._on_deserial(it_field, value))
                elif isinstance(it_field, ListField) and type(it_field.type.type) == Reference:
                    value = [self._deserial_reference_value(
                        it_field.type.type.model, it_obj, callbacks, key_to_instance)
                        for it_obj in v.list.value]
                    setattr(model_instance, it_field.attr, self._on_deserial(it_field, value))
                else:
                    setattr(model_instance, it_field.attr,
                            self._get_value(v, it_field.type, callbacks, it_field.weak_ref))
        else:
            for o_tag in d_obj:
                it_field = tag_to_fields[o_tag]
                it_field.deserialize(self, model_instance, AttrWrapper(d_obj),
                                     callbacks, key_to_instance)

        if isinstance(model_instance, Base):
            key_to_instance[model_instance.key, model_instance.id] = model_instance
        return model_instance



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
from functools import partial

import numpy as np
import pandas as pd

from ... import opcodes as OperandDef
from ...core import Entity, Base, OutputType
from ...lib.groupby_wrapper import wrapped_groupby
from ...operands import OperandStage
from ...serialize import BoolField, Int32Field, AnyField
from ...utils import get_shuffle_input_keys_idxes
from ..align import align_dataframe_series, align_series_series
from ..initializer import Series as asseries
from ..core import SERIES_TYPE, SERIES_CHUNK_TYPE
from ..utils import build_concatenated_rows_frame, hash_dataframe_on, \
    build_df, build_series, parse_index
from ..operands import DataFrameOperandMixin, DataFrameMapReduceOperand, DataFrameShuffleProxy


class DataFrameGroupByOperand(DataFrameMapReduceOperand, DataFrameOperandMixin):
    _op_type_ = OperandDef.GROUPBY

    _by = AnyField('by', on_serialize=lambda x: x.data if isinstance(x, Entity) else x)
    _level = AnyField('level')
    _as_index = BoolField('as_index')
    _sort = BoolField('sort')
    _group_keys = BoolField('group_keys')

    _shuffle_size = Int32Field('shuffle_size')

    def __init__(self, by=None, level=None, as_index=None, sort=None, group_keys=None,
                 shuffle_size=None, stage=None, shuffle_key=None, output_types=None, **kw):
        super().__init__(_by=by, _level=level, _as_index=as_index, _sort=sort,
                         _group_keys=group_keys, _shuffle_size=shuffle_size, _stage=stage,
                         _shuffle_key=shuffle_key, _output_types=output_types, **kw)
        if output_types:
            if stage in (OperandStage.map, OperandStage.reduce):
                if output_types[0] in (OutputType.dataframe, OutputType.dataframe_groupby):
                    output_types = [OutputType.dataframe]
                else:
                    output_types = [OutputType.series]
            else:
                if output_types[0] in (OutputType.dataframe, OutputType.dataframe_groupby):
                    output_types = [OutputType.dataframe_groupby]
                elif output_types[0] == OutputType.series:
                    output_types = [OutputType.series_groupby]
            self.output_types = output_types

    @property
    def by(self):
        return self._by

    @property
    def level(self):
        return self._level

    @property
    def as_index(self):
        return self._as_index

    @property
    def sort(self):
        return self._sort

    @property
    def group_keys(self):
        return self._group_keys

    @property
    def shuffle_size(self):
        return self._shuffle_size

    @property
    def is_dataframe_obj(self):
        return self.output_types[0] in (OutputType.dataframe_groupby, OutputType.dataframe)

    @property
    def groupby_params(self):
        return dict(by=self.by, level=self.level, as_index=self.as_index, sort=self.sort,
                    group_keys=self.group_keys)

    def build_mock_groupby(self, **kwargs):
        in_df = self.inputs[0]
        if self.is_dataframe_obj:
            empty_df = build_df(in_df, size=2)
            obj_dtypes = in_df.dtypes[in_df.dtypes == np.dtype('O')]
            empty_df[obj_dtypes.index] = 'O'
        else:
            if in_df.dtype == np.dtype('O'):
                empty_df = pd.Series('O', index=pd.RangeIndex(2), name=in_df.name, dtype=np.dtype('O'))
            else:
                empty_df = build_series(in_df, size=2, name=in_df.name)

        new_kw = self.groupby_params
        new_kw.update(kwargs)
        if new_kw.get('level'):
            new_kw['level'] = 0
        if isinstance(new_kw['by'], list):
            new_by = []
            for v in new_kw['by']:
                if isinstance(v, (Base, Entity)):
                    new_by.append(build_series(v, size=2, name=v.name))
                else:
                    new_by.append(v)
            new_kw['by'] = new_by
        return empty_df.groupby(**new_kw)

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        inputs_iter = iter(self._inputs[1:])
        if len(inputs) > 1:
            by = []
            for k in self._by:
                if isinstance(k, (SERIES_TYPE, SERIES_CHUNK_TYPE)):
                    by.append(next(inputs_iter))
                else:
                    by.append(k)
            self._by = by

    def __call__(self, df):
        params = df.params.copy()
        params['index_value'] = parse_index(None, df.key, df.index_value.key)
        if df.ndim == 2:
            if isinstance(self.by, list):
                index, types = [], []
                for k in self.by:
                    if isinstance(k, str):
                        if k not in df.dtypes:
                            raise KeyError(k)
                        else:
                            index.append(k)
                            types.append(df.dtypes[k])
                    else:
                        assert isinstance(k, SERIES_TYPE)
                        index.append(k.name)
                        types.append(k.dtype)
                params['key_dtypes'] = pd.Series(types, index=index)

        inputs = [df]
        if isinstance(self.by, list):
            for k in self.by:
                if isinstance(k, SERIES_TYPE):
                    inputs.append(k)

        return self.new_tileable(inputs, **params)

    @classmethod
    def _align_input_and_by(cls, op, inp, by):
        align_method = partial(align_dataframe_series, axis='index') \
            if op.is_dataframe_obj else align_series_series
        nsplits, _, inp_chunks, by_chunks = align_method(inp, by)

        inp_params = inp.params
        inp_params['chunks'] = inp_chunks
        inp_params['nsplits'] = nsplits
        inp = op.copy().new_tileable(op.inputs, kws=[inp_params])

        by_params = by.params
        by_params['chunks'] = by_chunks
        if len(nsplits) == 2:
            by_nsplits = nsplits[:1]
        else:
            by_nsplits = nsplits
        by_params['nsplits'] = by_nsplits
        by = by.op.copy().new_tileable(by.op.inputs, kws=[by_params])

        return inp, by

    @classmethod
    def tile(cls, op):
        in_df = op.inputs[0]
        by = op.by

        series_in_by = False
        new_inputs = []
        if len(op.inputs) > 1:
            # by series
            new_by = []
            for k in by:
                if isinstance(k, SERIES_TYPE):
                    in_df, k = cls._align_input_and_by(op, in_df, k)
                    if len(new_inputs) == 0:
                        new_inputs.append(in_df)
                    new_inputs.append(k)
                    series_in_by = True
                new_by.append(k)
            by = new_by
        else:
            new_inputs = op.inputs

        is_dataframe_obj = op.is_dataframe_obj
        if is_dataframe_obj:
            in_df = build_concatenated_rows_frame(in_df)
            output_type = OutputType.dataframe
            chunk_shape = (in_df.chunk_shape[0], 1)
        else:
            output_type = OutputType.series
            chunk_shape = (in_df.chunk_shape[0],)

        # generate map chunks
        map_chunks = []
        for chunk in in_df.chunks:
            map_op = op.copy().reset_key()
            map_op._stage = OperandStage.map
            map_op._shuffle_size = chunk_shape[0]
            chunk_inputs = [chunk]
            if len(op.inputs) > 1:
                chunk_by = []
                for k in by:
                    if isinstance(k, SERIES_TYPE):
                        by_chunk = k.cix[chunk.index[0], ]
                        chunk_by.append(by_chunk)
                        chunk_inputs.append(by_chunk)
                    else:
                        chunk_by.append(k)
                map_op._by = chunk_by
            map_chunks.append(map_op.new_chunk(chunk_inputs, shape=(np.nan, np.nan), index=chunk.index))

        proxy_chunk = DataFrameShuffleProxy(output_types=[output_type]).new_chunk(map_chunks, shape=())

        # generate reduce chunks
        reduce_chunks = []
        for out_idx in itertools.product(*(range(s) for s in chunk_shape)):
            shuffle_key = ','.join(str(idx) for idx in out_idx)
            reduce_op = op.copy().reset_key()
            reduce_op._by = None
            reduce_op._stage = OperandStage.reduce
            reduce_op._shuffle_key = shuffle_key
            reduce_chunks.append(
                reduce_op.new_chunk([proxy_chunk], shape=(np.nan, np.nan), index=out_idx))

        # generate groupby chunks
        out_chunks = []
        for chunk in reduce_chunks:
            groupby_op = op.copy().reset_key()
            if series_in_by:
                # set by to None, cuz data of by will be passed from map to reduce to groupby
                groupby_op._by = None
            if is_dataframe_obj:
                new_shape = (np.nan, in_df.shape[1])
            else:
                new_shape = (np.nan,)
            params = dict(shape=new_shape, index=chunk.index)
            if op.is_dataframe_obj:
                params.update(dict(dtypes=in_df.dtypes, columns_value=in_df.columns_value,
                                   index_value=parse_index(None, chunk.key, proxy_chunk.key)))
            else:
                params.update(dict(name=in_df.name, dtype=in_df.dtype,
                                   index_value=parse_index(None, chunk.key, proxy_chunk.key)))
            out_chunks.append(groupby_op.new_chunk([chunk], **params))

        new_op = op.copy()
        params = op.outputs[0].params.copy()
        if is_dataframe_obj:
            params['nsplits'] = ((np.nan,) * len(out_chunks), (in_df.shape[1],))
        else:
            params['nsplits'] = ((np.nan,) * len(out_chunks),)
        params['chunks'] = out_chunks
        return new_op.new_tileables(new_inputs, **params)

    @classmethod
    def execute_map(cls, ctx, op):
        is_dataframe_obj = op.is_dataframe_obj
        by = op.by
        chunk = op.outputs[0]
        df = ctx[op.inputs[0].key]

        deliver_by = False  # output by for the upcoming process
        if isinstance(by, list):
            new_by = []
            for v in by:
                if isinstance(v, Base):
                    deliver_by = True
                    new_by.append(ctx[v.key])
                else:
                    new_by.append(v)
            by = new_by

        if isinstance(by, list) or callable(by):
            on = by
        else:
            on = None
        filters = hash_dataframe_on(df, on, op.shuffle_size, level=op.level)

        for index_idx, index_filter in enumerate(filters):
            if is_dataframe_obj:
                group_key = ','.join([str(index_idx), str(chunk.index[1])])
            else:
                group_key = str(index_idx)

            if deliver_by:
                filtered_by = []
                for v in by:
                    if isinstance(v, pd.Series):
                        filtered_by.append(v.loc[index_filter])
                    else:
                        filtered_by.append(v)
                ctx[(chunk.key, group_key)] = (df.loc[index_filter], filtered_by)
            else:
                ctx[(chunk.key, group_key)] = df.loc[index_filter]

    @classmethod
    def execute_reduce(cls, ctx, op):
        is_dataframe_obj = op.inputs[0].op.output_types[0] == OutputType.dataframe
        chunk = op.outputs[0]
        input_keys, input_idxes = get_shuffle_input_keys_idxes(op.inputs[0])
        input_idx_to_df = {idx: ctx[inp_key, ','.join(str(ix) for ix in chunk.index)]
                           for inp_key, idx in zip(input_keys, input_idxes)}
        row_idxes = sorted({idx[0] for idx in input_idx_to_df})

        res = []
        for row_idx in row_idxes:
            if is_dataframe_obj:
                row_df = input_idx_to_df.get((row_idx, 0), None)
            else:
                row_df = input_idx_to_df.get((row_idx,), None)
            if row_df is not None:
                res.append(row_df)
        by = None
        if isinstance(res[0], tuple):
            # By is series
            r = pd.concat([it[0] for it in res], axis=0)
            by = [None] * len(res[0][1])
            for it in res:
                for i, v in enumerate(it[1]):
                    if isinstance(v, pd.Series):
                        if by[i] is None:
                            by[i] = v
                        else:
                            by[i] = pd.concat([by[i], v], axis=0)
                    else:
                        by[i] = v
        else:
            r = pd.concat(res, axis=0)
        if chunk.index_value is not None:
            r.index.name = chunk.index_value.name
        if by is None:
            ctx[chunk.key] = r
        else:
            ctx[chunk.key] = (r, by)

    @classmethod
    def execute(cls, ctx, op: "DataFrameGroupByOperand"):
        if op.stage == OperandStage.map:
            cls.execute_map(ctx, op)
        elif op.stage == OperandStage.reduce:
            cls.execute_reduce(ctx, op)
        else:
            inp = ctx[op.inputs[0].key]
            if isinstance(inp, tuple):
                # df, by
                df, by = inp
            else:
                df = inp
                by = op.by
            ctx[op.outputs[0].key] = wrapped_groupby(
                df, by=by, level=op.level, as_index=op.as_index, sort=op.sort,
                group_keys=op.group_keys)


def groupby(df, by=None, level=None, as_index=True, sort=True, group_keys=True):
    if not as_index and df.op.output_types[0] == OutputType.series:
        raise TypeError('as_index=False only valid with DataFrame')

    output_types = [OutputType.dataframe_groupby] if df.ndim == 2 else [OutputType.series_groupby]
    if isinstance(by, (str, SERIES_TYPE, pd.Series)):
        if isinstance(by, pd.Series):
            by = asseries(by)
        by = [by]
    op = DataFrameGroupByOperand(by=by, level=level, as_index=as_index, sort=sort,
                                 group_keys=group_keys, output_types=output_types)
    return op(df)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ...utils import copy_tileables
from ...dataframe.utils import parse_index
from ...dataframe.groupby.aggregation import DataFrameGroupByAgg
from ...dataframe.datasource.read_csv import DataFrameReadCSV
from .core import TileableOptimizeRule, register


class GroupbyPruneReadCSV(TileableOptimizeRule):
    """
    An experimental implementation for tileable optimization rule.
    This rule works only when groupby aggregation operation follows the read CSV files,
    we can prune the columns that not used by the following operations when read the files.
    """
    def match(self, node):
        if isinstance(node.inputs[0].op, DataFrameReadCSV) and \
                node.inputs[0] not in self._optimizer_context.result_tileables:
            by_data = node.op.groupby_params.get('by')
            by_columns = by_data if isinstance(by_data, (list, tuple)) else [by_data]
            if isinstance(node.op.func, (str, list)):
                # Passing func name(s) means perform on all columns.
                return False
            elif len(set(by_columns + list(node.op.func))) == len(node.inputs[0].op.usecols or node.inputs[0].dtypes):
                # If performs on all columns, no need to prune.
                return False
            return True
        return False

    def apply(self, node):
        by_data = node.op.groupby_params.get('by')
        by_columns = by_data if isinstance(by_data, (list, tuple)) else [by_data]
        agg_columns = list(node.op.func)
        input_node = node.inputs[0]
        selected_columns = [c for c in list(input_node.dtypes.index) if c in by_columns + agg_columns]
        if input_node in self._optimizer_context:
            new_input = self._optimizer_context[input_node]
            selected_columns = [
                c for c in list(input_node.dtypes.index) if c in selected_columns + new_input.op._usecols]
        else:
            new_input = copy_tileables([input_node])[0].data

        new_input._shape = (input_node.shape[0], len(selected_columns))
        new_input._dtypes = input_node.dtypes[selected_columns]
        new_input._columns_value = parse_index(new_input._dtypes.index, store_data=True)
        new_input.op._usecols = selected_columns
        new_node = copy_tileables([node], inputs=[new_input])[0].data

        self._optimizer_context[node] = new_node
        self._optimizer_context[input_node] = new_input
        return node


register(DataFrameGroupByAgg, GroupbyPruneReadCSV)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ...serialize import KeyField, TupleField
from ..operands import TensorHasInput, TensorOperandMixin
from ..datasource import tensor as astensor
from ..array_utils import get_array_module, device


class TensorBroadcastTo(TensorHasInput, TensorOperandMixin):
    _op_type_ = OperandDef.BROADCAST_TO

    _input = KeyField('input')
    _shape = TupleField('shape')

    def __init__(self, shape=None, dtype=None, sparse=False, **kw):
        super().__init__(_shape=shape, _dtype=dtype, _sparse=sparse, **kw)

    @property
    def shape(self):
        return self._shape

    def __call__(self, tensor, shape):
        return self.new_tensor([tensor], shape)

    @classmethod
    def tile(cls, op):
        tensor = op.outputs[0]
        in_tensor = op.inputs[0]
        shape = op.shape
        new_dim = tensor.ndim - in_tensor.ndim

        out_chunks = []
        for c in in_tensor.chunks:
            chunk_shape = shape[:new_dim] + tuple(s if in_tensor.shape[idx] != 1 else shape[new_dim+idx]
                                                  for idx, s in enumerate(c.shape))
            chunk_idx = (0,) * new_dim + c.index
            chunk_op = op.copy().reset_key()
            chunk_op._shape = chunk_shape
            out_chunk = chunk_op.new_chunk([c], shape=chunk_shape, index=chunk_idx, order=tensor.order)
            out_chunks.append(out_chunk)

        nsplits = [tuple(c.shape[i] for c in out_chunks if all(idx == 0 for j, idx in enumerate(c.index) if j != i))
                   for i in range(len(out_chunks[0].shape))]
        new_op = op.copy()
        return new_op.new_tensors([in_tensor], tensor.shape, order=tensor.order,
                                  chunks=out_chunks, nsplits=nsplits)

    @classmethod
    def execute(cls, ctx, op):
        xp = get_array_module(ctx[op.input.key])
        input_data = ctx[op.input.key]
        device_id = input_data.device.id if hasattr(input_data, 'device') else -1

        with device(device_id):
            shape = op.shape
            if any(np.isnan(s) for s in shape):
                shape = list(shape)
                new_dim = len(shape) - input_data.ndim
                for i in range(input_data.ndim):
                    if np.isnan(shape[i + new_dim]):
                        shape[i + new_dim] = input_data.shape[i]
            ctx[op.outputs[0].key] = xp.broadcast_to(input_data, shape)


def broadcast_to(tensor, shape):
    """Broadcast an tensor to a new shape.

    Parameters
    ----------
    tensor : array_like
        The tensor to broadcast.
    shape : tuple
        The shape of the desired array.

    Returns
    -------
    broadcast : Tensor

    Raises
    ------
    ValueError
        If the tensor is not compatible with the new shape according to Mars's
        broadcasting rules.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> x = mt.array([1, 2, 3])
    >>> mt.broadcast_to(x, (3, 3)).execute()
    array([[1, 2, 3],
           [1, 2, 3],
           [1, 2, 3]])
    """
    from ..core import Tensor

    tensor = tensor if isinstance(tensor, Tensor) else astensor(tensor)
    shape = tuple(shape) if isinstance(shape, (list, tuple)) else (shape,)

    if any(np.isnan(s) for s in tensor.shape):
        raise ValueError('input tensor has unknown shape, '
                         'need to call `.execute()` first')

    if tensor.shape == shape:
        return tensor

    new_ndim = len(shape) - tensor.ndim
    if new_ndim < 0:
        raise ValueError('input operand has more dimensions than allowed by the axis remapping')
    if any(o != n for o, n in zip(tensor.shape, shape[new_ndim:]) if o != 1):
        raise ValueError('operands could not be broadcast together '
                         f'with remapped shapes [original->remapped]: {tensor.shape} '
                         f'and requested shape {shape}')

    op = TensorBroadcastTo(shape, dtype=tensor.dtype, sparse=tensor.issparse())
    return op(tensor, shape)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ...serialize import KeyField, AnyField, BoolField
from ...lib.sparse.core import get_array_module
from ...utils import check_chunks_unknown_shape
from ...tiles import TilesError
from ..array_utils import as_same_device, device
from ..core import Tensor, TensorOrder
from ..datasource import tensor as astensor
from ..operands import TensorHasInput, TensorOperandMixin


class TensorDigitize(TensorHasInput, TensorOperandMixin):
    _op_type_ = OperandDef.DIGITIZE

    _input = KeyField('input')
    _bins = AnyField('bins')
    _right = BoolField('right')

    def __init__(self, right=False, dtype=None, **kw):
        super().__init__(_right=right, _dtype=dtype, **kw)

    @property
    def bins(self):
        return self._bins

    @property
    def right(self):
        return self._right

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._input = self._inputs[0]
        if len(inputs) > 1:
            self._bins = self._inputs[1]

    def __call__(self, x, bins):
        x = astensor(x)
        inputs = [x]
        if not isinstance(bins, Tensor):
            bins = get_array_module(bins).asarray(bins)
            self._bins = bins
        else:
            inputs.append(bins)
        self._dtype = np.digitize([0], np.empty(1, dtype=bins.dtype), right=self._right).dtype

        return self.new_tensor(inputs, x.shape, order=TensorOrder.C_ORDER)

    @classmethod
    def tile(cls, op):
        tensor = op.outputs[0]
        in_tensor = op.input
        bins = op.bins
        if len(op.inputs) == 2:
            # bins is TensorData
            check_chunks_unknown_shape([bins], TilesError)
            bins = bins.rechunk(tensor.shape)._inplace_tile().chunks[0]

        out_chunks = []
        for c in in_tensor.chunks:
            input_chunks = [c]
            if len(op.inputs) == 2:
                input_chunks.append(bins)
            out_chunk = op.copy().reset_key().new_chunk(input_chunks, shape=c.shape,
                                                        index=c.index, order=tensor.order)
            out_chunks.append(out_chunk)

        new_op = op.copy()
        return new_op.new_tensors(op.inputs, tensor.shape, order=tensor.order,
                                  chunks=out_chunks, nsplits=in_tensor.nsplits)

    @classmethod
    def execute(cls, ctx, op):
        inputs, device_id, xp = as_same_device(
            [ctx[c.key] for c in op.inputs], device=op.device, ret_extra=True)

        x = inputs[0]
        if len(inputs) > 1:
            bins = inputs[1]
        else:
            bins = op.bins

        with device(device_id):
            ctx[op.outputs[0].key] = xp.digitize(x, bins=bins, right=op.right)


def digitize(x, bins, right=False):
    """
    Return the indices of the bins to which each value in input tensor belongs.

    Each index ``i`` returned is such that ``bins[i-1] <= x < bins[i]`` if
    `bins` is monotonically increasing, or ``bins[i-1] > x >= bins[i]`` if
    `bins` is monotonically decreasing. If values in `x` are beyond the
    bounds of `bins`, 0 or ``len(bins)`` is returned as appropriate. If right
    is True, then the right bin is closed so that the index ``i`` is such
    that ``bins[i-1] < x <= bins[i]`` or ``bins[i-1] >= x > bins[i]`` if `bins`
    is monotonically increasing or decreasing, respectively.

    Parameters
    ----------
    x : array_like
        Input tensor to be binned.
    bins : array_like
        Array of bins. It has to be 1-dimensional and monotonic.
    right : bool, optional
        Indicating whether the intervals include the right or the left bin
        edge. Default behavior is (right==False) indicating that the interval
        does not include the right edge. The left bin end is open in this
        case, i.e., bins[i-1] <= x < bins[i] is the default behavior for
        monotonically increasing bins.

    Returns
    -------
    out : Tensor of ints
        Output tensor of indices, of same shape as `x`.

    Raises
    ------
    ValueError
        If `bins` is not monotonic.
    TypeError
        If the type of the input is complex.

    See Also
    --------
    bincount, histogram, unique, searchsorted

    Notes
    -----
    If values in `x` are such that they fall outside the bin range,
    attempting to index `bins` with the indices that `digitize` returns
    will result in an IndexError.

    `mt.digitize` is  implemented in terms of `mt.searchsorted`. This means
    that a binary search is used to bin the values, which scales much better
    for larger number of bins than the previous linear search. It also removes
    the requirement for the input array to be 1-dimensional.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> x = mt.array([0.2, 6.4, 3.0, 1.6])
    >>> bins = mt.array([0.0, 1.0, 2.5, 4.0, 10.0])
    >>> inds = mt.digitize(x, bins)
    >>> inds.execute()
    array([1, 4, 3, 2])

    >>> x = mt.array([1.2, 10.0, 12.4, 15.5, 20.])
    >>> bins = mt.array([0, 5, 10, 15, 20])
    >>> mt.digitize(x,bins,right=True).execute()
    array([1, 2, 3, 4, 4])
    >>> mt.digitize(x,bins,right=False).execute()
    array([1, 3, 3, 4, 5])
    """
    op = TensorDigitize(right=right)
    return op(x, bins)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

try:
    from scipy.special import entr as scipy_entr
except ImportError:  # pragma: no cover
    scipy_entr = None

from ... import opcodes as OperandDef
from ...utils import require_not_none
from ..arithmetic.utils import arithmetic_operand
from ..utils import infer_dtype
from .core import TensorSpecialUnaryOp


@arithmetic_operand(sparse_mode='unary')
class TensorEntr(TensorSpecialUnaryOp):
    _op_type_ = OperandDef.ENTR
    _func_name = 'entr'


@require_not_none(scipy_entr)
@infer_dtype(scipy_entr)
def entr(x, out=None, where=None, **kwargs):
    r"""
    Elementwise function for computing entropy.

    .. math:: \text{entr}(x) = \begin{cases} - x \log(x) & x > 0  \\ 0 & x = 0 \\ -\infty & \text{otherwise} \end{cases}

    Parameters
    ----------
    x : Tensor
        Input tensor.

    Returns
    -------
    res : Tensor
        The value of the elementwise entropy function at the given points `x`.

    See Also
    --------
    kl_div, rel_entr

    Notes
    -----
    This function is concave.
    """
    op = TensorEntr(**kwargs)
    return op(x, out=out, where=where)
